import json
import numpy as np
from scipy.interpolate import BSpline

# Load model data from JSON
with open('gam_model.json', 'r') as file:
    model_data = json.load(file)

# Extract model parameters from loaded JSON
coefficients = np.array(model_data['coefficients'])
RBP_Length_knots = np.array(model_data['RBP_Length_knots'], dtype=float)  # Ensure floats
TBD_length_knots = np.array(model_data['TBD_length_knots'], dtype=float)  # Ensure floats

# Define prediction function in Python
def predict_gam_model(RBP_Length, TBD_length):
    # Implement prediction logic using coefficients and knots
    linear_pred = np.dot(coefficients, [1, RBP_Length, TBD_length, RBP_Length**2, TBD_length**2])  # Example linear prediction
    spline_pred = BSpline(RBP_Length_knots, np.zeros_like(RBP_Length_knots), 3)(TBD_length)
    return linear_pred + spline_pred

# Example usage
predicted_rate = predict_gam_model(6, 50)
print(f"Predicted Mutation Rate: {predicted_rate}")



#!/usr/bin/env python

import unittest
import os
import subprocess
import csv
import shutil
import pathlib
import efmcalculator
import Bio.SeqIO as SeqIO
import polars as pl
import efmcalculator
from itertools import chain

from efmcalculator.ingest.bad_state_mitigation import detect_special_cases

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
output_directory_path = os.path.join(THIS_DIR)


def csv_are_identical(csv_file_path_1, csv_file_path_2):
    'Utility function for comparing output, Returns whether files are identical as bool'
    failed = False
    try:
        with open(csv_file_path_1) as csv1:
            reader = csv.reader(csv1)
            csv_1_contents = [row for row in reader]
    except:
        print(f'Failed to open {str(csv_file_path_1)}')
        failed = True

    try:
        with open(csv_file_path_2) as csv2:
            reader = csv.reader(csv2)
            csv_2_contents = [row for row in reader]
    except:
        print(f'Failed to open {str(csv_file_path_2)}')
        failed = True

    if len(csv_1_contents) != len(csv_2_contents):
        print(f'CSVs are different lengths')
        failed = True

    differences = []
    for i in range(0, len(csv_1_contents)):
        if csv_1_contents[i] != csv_2_contents[i]:
            # print(csv_1_rows[i], "\n!=\n",csv_1_rows[i])
            differences.append(i)
    if differences:
        print(f'CSVs have {len(differences)} different values at rows {differences}')
        failed = True

    if failed == True:
        return False
    else:
        return True

##############################################################################################
# Unit tests (function calls)
##############################################################################################
class test_unit_run_complete_wo_errors(unittest.TestCase):
    """Tests to ensure that EFM2 is capable of completing a single example. Fails on error only"""
    def test_unit_run_complete_wo_errors(self):
        print("Test: Run complete without errors")
        L6_10_plasmid = output_directory_path  + "/../examples/1_L6-10_plasmid.gb"
        L6_10_plasmid = pathlib.Path(L6_10_plasmid)
        L6_10_plasmid = efmcalculator.ingest.parse_file(L6_10_plasmid)
        inseqs = list()
        statemachine = efmcalculator.StateMachine.StateMachine()
        statemachine.import_sequences(inseqs)
        for seqobject in statemachine.user_sequences.values():
            seqobject.call_predictions()


class test_unit_filtering_ssr(unittest.TestCase):
    """Tests SSR filtering. Fails when an SSR doesn't get filtered appropriately"""
    def test_unit_filtering_ssr(self):
        return
        test_sequence = "CACACACACA"+"ATGTTTTCACACACACA"*6 +"ATGTTTTGGGAGAGAGCACACAATGTTTT"
        shouldnt_exist = ["ACACACACAATGTTTTC",
                          "CACACACAATGTTTTCA",
                          "ACACACAATGTTTTCAC",
                          "CACACAATGTTTTCACA",
                          "ACACAATGTTTTCACAC",
                          "CACAATGTTTTCACACA",
                          "ACAATGTTTTCACACAC",
                          "CAATGTTTTCACACACA",
                          "AATGTTTTCACACACAC",
                          "ATGTTTTCACACACACA",
                          "AC",
                          "A",
                          "C",
                          "G",
                          "AG"]
        should_exist = ["CACACACACAATGTTTT"]

        for is_circular in [True, False]:
            if not is_circular:
                should_exist += ["T", "CA"]

            ssr_df, _, _ = efmcalculator.predict(test_sequence, "pairwise", is_circular)
            ssr_df = efmcalculator.pipeline.filtering.filter_ssrs(ssr_df, len(test_sequence), is_circular)

            ssr_df = ssr_df.with_columns(
                pl.lit(shouldnt_exist).alias("test_case")
            ).with_columns(
                pl.col("repeat").is_in(pl.col("test_case")).alias("fail")
            )
            #assert (True,) not in ssr_df.select("fail").unique().rows(), "SSR failing to filter some nested sequences"
            for positive_check in should_exist:
                detected_ssrs = ssr_df.filter(pl.col("repeat") == positive_check)
                assert detected_ssrs.height >= 1, "Known SSR is missing"

class test_unit_filtering_rmd(unittest.TestCase):
    def test_unit_filtering_rmd(self):
        test_sequence = "AGTCCAA" + "AAAAAAAAAAAAAAAAAAAAA" + "AGTCCAA"
        is_circular = False
        ssr_df, srs_df, rmd_df = efmcalculator.predict(test_sequence, "pairwise", is_circular)

if __name__ == "__main__":
    unittest.main()



import pathlib
import os
from .ingest.parse_inputs import validate_sequences
from .constants import MAX_SIZE
import polars as pl

import concurrent.futures as cf
from progress.bar import Bar
import threading
import multiprocessing as mp
from .pipeline.mutation_rates import rip_score
from .webapp.SequenceState import SequenceState

class ThreadSafeBar(Bar):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._lock = threading.Lock()

    def next(self):
        with self._lock:
            super().next()

def _parallel_worker(args):
    """Internal function for parallelizing sequence analysis across many sequences"""
    seqobj, seqname, strategy, keepmem = args
    seqobj.call_predictions(strategy)
    scores = rip_score(ssr_df=seqobj.ssrs, srs_df=seqobj.srss,
                        rmd_df=seqobj.rmds, sequence_length=len(seqobj))
    scores['name'] = seqname
    reorder = ['name', 'ssr_sum', 'srs_sum', 'rmd_sum', 'bps_sum', 'rip']
    scores = {k: scores[k] for k in reorder}
    sample_df = pl.DataFrame(scores).cast({
        'ssr_sum': pl.Float64,
        'srs_sum': pl.Float64,
        'rmd_sum': pl.Float64,
        'bps_sum': pl.Float64,
        'rip': pl.Float64
    })
    if not keepmem:
        seqobj = None
    return seqobj, scores

class StateMachine:
    """Class for recording user session state between streamlit interactions to prevent rerunning analysis and make
    selected predictions persist."""
    def __init__(self):
        self.user_sequences = {}
        self.named_sequences = {}
        self.sequencestates = {}


    def import_sequences(self, sequences, max_size=None, webapp = False):
        """Import newly uploaded sequences while retaining state of existing sequences"""
        # Import sequences without overwriting old ones
        new = {seq._originhash: seq for seq in sequences}
        for key in new:
            if key in self.user_sequences:
                new[key] = self.user_sequences[key]
        if new == self.user_sequences:
            return
        self.user_sequences = new

        # Validate sequences if they changed
        validate_sequences(self.user_sequences.values())

        # Make webapp states
        if webapp:
            self.sequencestates = {key: SequenceState(value) for key, value in self.user_sequences.items()}

        # Update sequence names
        self.named_sequences = {}
        for i, seqhash in enumerate(self.user_sequences):
            seq = self.user_sequences[seqhash]
            if seq.description:
                sequence_name = f"{i+1}_{seq.description}"
            else:
                sequence_name = f"{i+1}_Sequence"
            self.named_sequences[sequence_name] = seqhash

    def predict_tall(self, outpath, strategy, filetype, threads, keepmem=False, summaryonly=False):
        samples = []
        for seqname in self.named_sequences:
            seqhash = self.named_sequences[seqname]
            seqobj = self.user_sequences[seqhash]
            sample = (seqobj, seqname, strategy, keepmem)
            samples.append(sample)

        parallel_bar = ThreadSafeBar("Predicting hotspots", max=len(samples))

        mpcontext = mp.get_context("spawn")
        with cf.ProcessPoolExecutor(max_workers=threads, mp_context=mpcontext) as executor:
            results = [executor.submit(_parallel_worker, sample) for sample in samples]
            for f in cf.as_completed(results):
                parallel_bar.next()
        parallel_bar.finish()
        results = [result.result() for result in results]

        if keepmem:
            self.user_sequences = {result[0]._originhash: result[0] for result in results}
            self.save_results(outpath, filetype = filetype)
        else:
            results = [result[1] for result in results]
            results = {key: [d[key] for d in results] for key in results[0]}
            summary_df = pl.DataFrame(results, schema = {
                'name': pl.String,
                'ssr_sum': pl.Float64,
                'srs_sum': pl.Float64,
                'rmd_sum': pl.Float64,
                'bps_sum': pl.Float64,
                'rip': pl.Float64
            })
            if filetype == "parquet":
                summarypath = os.path.join(outpath, "summary.parquet")
                summary_df.write_parquet(summarypath)
            elif filetype == "csv":
                summarypath = os.path.join(outpath, "summary.csv")
                summary_df.write_csv(summarypath)

    def save_results(self, folderpath, prediction_style = None, filetype = "parquet", summaryonly=False):
        summary_df = pl.DataFrame([
            pl.Series("name", [], dtype=pl.String),
            pl.Series("ssr_sum", [], dtype=pl.Float64),
            pl.Series("srs_sum", [], dtype=pl.Float64),
            pl.Series("rmd_sum", [], dtype=pl.Float64),
            pl.Series("bps_sum", [], dtype=pl.Float64),
            pl.Series("rip", [], dtype=pl.Float64),
        ])
        for seqname in self.named_sequences:
            seqhash = self.named_sequences[seqname]
            seqobj = self.user_sequences[seqhash]
            if not seqobj.predicted:
                if not prediction_style:
                    raise ValueError("Must specify prediction style to save results")
                elif prediction_style not in ['linear', 'pairwise']:
                    raise ValueError("Invalid prediction style: linear or pairwise")
                seqobj.call_predictions(prediction_style)

            scores = rip_score(ssr_df=seqobj.ssrs, srs_df=seqobj.srss,
                                rmd_df=seqobj.rmds, sequence_length=len(seqobj))
            scores['name'] = seqname
            reorder = ['name', 'ssr_sum', 'srs_sum', 'rmd_sum', 'bps_sum', 'rip']
            scores = {k: scores[k] for k in reorder}
            sample_df = pl.DataFrame(scores).cast({
                'ssr_sum': pl.Float64,
                'srs_sum': pl.Float64,
                'rmd_sum': pl.Float64,
                'bps_sum': pl.Float64,
                'rip': pl.Float64
            })
            summary_df.extend(sample_df)

            if summaryonly:
                continue

            top = seqobj.top.select(pl.exclude("predid"))
            ssrs = seqobj.ssrs.select(pl.exclude(["predid", "annotationobjects"]))
            srss = seqobj.srss.select(pl.exclude(["predid", "annotationobjects"]))
            rmds = seqobj.rmds.select(pl.exclude(["predid", "annotationobjects"]))

            folder = os.path.join(folderpath, f"{seqname}")
            path = pathlib.Path(folder)
            path.mkdir(parents=True)
            if filetype == "parquet":
                top.write_parquet(os.path.join(folder, "top.parquet"))
                ssrs.write_parquet(os.path.join(folder, "ssrs.parquet"))
                srss.write_parquet(os.path.join(folder, "srss.parquet"))
                rmds.write_parquet(os.path.join(folder, "rmds.parquet"))
            elif filetype == "csv":
                top.select(pl.exclude("annotations")).write_csv(os.path.join(folder, "top.csv"))
                ssrs.select(pl.exclude("annotations")).write_csv(os.path.join(folder, "ssrs.csv"))
                srss.select(pl.exclude("annotations")).write_csv(os.path.join(folder, "srss.csv"))
                rmds.select(pl.exclude("annotations")).write_csv(os.path.join(folder, "rmds.csv"))
            else:
                raise ValueError("Invalid filetype")

        if filetype == "parquet":
            summarypath = os.path.join(folderpath, "summary.parquet")
            summary_df.write_parquet(summarypath)
        elif filetype == "csv":
            summarypath = os.path.join(folderpath, "summary.csv")
            summary_df.write_csv(summarypath)



VALID_STRATEGIES = ["pairwise", "linear"]
MIN_SHORT_SEQ_LEN = 0
MAX_SHORT_SEQ_LEN = 16
MIN_SRS_LEN = 6 # Dont search for SRSs below this length
MAX_SSR_SPACER = 5000 # Dont search for SRSs this far appart
UNKNOWN_REC_TYPE = "unknown"
FASTA_EXTS = [".fa", ".fasta"]
GBK_EXTS = [".gb", ".gbk", ".gbff"]
CSV_EXTS = [".csv"]
VALID_EXTS = FASTA_EXTS + GBK_EXTS + CSV_EXTS
SUB_RATE = float(2.2 * 10 ** (-10))
THRESHOLD = 2.2*10e-10 # Minimum mutation rate for a reported hotspot
MAX_SIZE = 50000
MARKER_HEIGHT = 500
VALID_FILETYPES = ["csv", "parquet"]

COLORS = {
    "ori": "#4e7fff",
    "promoter": "#f6a35e",
    "cds": "#479f71",
    "misc": "#808080",
    "primer_bind": "#d0d0d0",
    "terminator": "#C97064",
    "ncrna": "#E8DAB2",
}



from .mutation_rates import ssr_mut_rate_vector, srs_mut_rate_vector, rmd_mut_rate_vector
from .filtering import filter_ssrs, filter_direct_repeats
from .features import assign_features_ssr, assign_features_rmd
from ..constants import THRESHOLD

import polars as pl

def post_process(ssr_df, srs_df, rmd_df, seqobj, isCircular):
    # Perform Filtering
    ssr_df = filter_ssrs(ssr_df, len(seqobj), isCircular)
    rmd_df, srs_df = filter_direct_repeats(rmd_df, srs_df, len(seqobj), ssr_df, isCircular)

    # Calculate Mutation Rates

    ssr_df = ssr_mut_rate_vector(ssr_df)
    srs_df = srs_mut_rate_vector(srs_df)
    rmd_df = rmd_mut_rate_vector(rmd_df)

    # Filter on minimum threshold
    ssr_df = ssr_df.filter(pl.col("mutation_rate") > THRESHOLD)
    srs_df = srs_df.filter(pl.col("mutation_rate") > THRESHOLD)
    rmd_df = rmd_df.filter(pl.col("mutation_rate") > THRESHOLD)

    # Apply annotations
    if seqobj.annotations:
        ssr_df = assign_features_ssr(ssr_df, seqobj, isCircular)
        srs_df = assign_features_rmd(srs_df, seqobj, isCircular)
        rmd_df = assign_features_rmd(rmd_df, seqobj, isCircular)

    return ssr_df, srs_df, rmd_df



import pathlib
import logging
import csv
from Bio import SeqIO, SeqRecord, Seq
from Bio.SeqFeature import SimpleLocation, CompoundLocation, SeqFeature
from pathlib import Path
from polars import DataFrame
import polars as pl

def seqfeature_hash(seqfeature: SeqFeature):
    """Hash for a SeqFeature objects"""
    return hash(str(seqfeature))

def assign_features_ssr(ssrdf: DataFrame, seqobj: Seq, circular: bool):
    if ssrdf.is_empty():
        ssrdf = ssrdf.with_columns(pl.lit(None).alias("annotations").cast(pl.List(pl.String)),
            pl.lit(None).alias("annotationobjects").cast(pl.List(pl.Int64)))
        return ssrdf

    # Test to see if dataframe has a position and length column
    features = sequence_to_features_df(seqobj, circular)
    ssrdf = ssrdf.with_columns((pl.col("start") + pl.col("repeat_len")*pl.col("count")-1).alias("end").cast(pl.Int32))

    # Annotation on Left edge
    left_edge = ssrdf.join_where(features, (pl.col("left_bound") <= pl.col("start"))
                                         & (pl.col("start") <= pl.col("right_bound"))
                                         & (pl.col("end") > pl.col("right_bound")))
    left_edge = left_edge.with_columns(pl.lit("edge").alias("featureclass"))

    # Annotation on right edge
    right_edge = ssrdf.join_where(features, (pl.col("left_bound") <= pl.col("end"))
                                         & (pl.col("end") <= pl.col("right_bound"))
                                         & (pl.col("start") < pl.col("left_bound")))
    right_edge = right_edge.with_columns(pl.lit("edge").alias("featureclass"))

    # Annotation inside
    inside = ssrdf.join_where(features, (pl.col("start") <= pl.col("left_bound")) & (pl.col("end") >= pl.col("right_bound")))
    inside = inside.with_columns(pl.lit("inside").alias("featureclass"))

    # Annotation wraps
    wraps = ssrdf.join_where(features, (pl.col("start") >= pl.col("left_bound")) & (pl.col("end") <= pl.col("right_bound")))
    wraps = wraps.with_columns(pl.lit("wraps").alias("featureclass"))

    anno = pl.concat([left_edge, right_edge, inside, wraps]).unique()
    anno = anno.group_by(["repeat", "repeat_len", "start", "count", "mutation_rate"]
                        ).agg(pl.col('annotations'), pl.col('annotationobjects')
    )

    intergenic = ssrdf.join(anno.select(["repeat", "repeat_len", "start", "count"]), how="anti", on=["repeat", "repeat_len", "start", "count"]
    ).with_columns(pl.lit([]).alias("annotations")).with_columns(pl.lit([]).alias("annotationobjects")).select(["repeat", "repeat_len", "start", "count", "mutation_rate", "annotations", "annotationobjects"])


    ssrdf = pl.concat([anno, intergenic])

    return ssrdf


def assign_features_rmd(rmd_or_srs_df: DataFrame, seqobj: Seq, circular: bool):

    if rmd_or_srs_df.is_empty():
        rmd_or_srs_df = rmd_or_srs_df.with_columns(pl.lit(None).alias("annotations").cast(pl.List(pl.String)),
            pl.lit(None).alias("annotationobjects").cast(pl.List(pl.Int64)))
        return rmd_or_srs_df

    # Test to see if a dataframe has a position left, position right, and length column
    features = sequence_to_features_df(seqobj, circular)
    pl.Config.set_tbl_rows(100)
    sequence_length = len(seqobj)
    df = (rmd_or_srs_df
            .with_columns(pl.when(
            pl.col("second_repeat")-pl.col("first_repeat")-pl.col("repeat_len") != pl.col("distance")
            ).then(pl.lit(True)
            ).otherwise(pl.lit(False)
            ).alias("wraps")))

    df =    (df.with_columns(pl.when(
            pl.col("wraps") == False
            ).then(pl.col("first_repeat"))
            .otherwise(pl.col("second_repeat")).alias("start")
            )
            .with_columns(
            pl.when(
                pl.col("wraps") == False
            ).then(pl.col("second_repeat") + pl.col("repeat_len"))
            .otherwise(pl.col("first_repeat") + pl.col("repeat_len")).alias("end")
            ))

    # Annotation on Left edge
    left_edge = df.filter(pl.col("wraps") == False).join_where(features, ((pl.col("left_bound") <= pl.col("start")))
                                            & (pl.col("start") <= pl.col("right_bound"))
                                            & (pl.col("end") > pl.col("right_bound"))
                                            )
    left_edge_wraps = df.filter(pl.col("wraps") == True).join_where(features, (
                                            pl.col("left_bound") <= pl.col("start"))
                                            & (pl.col("right_bound") >= pl.col("start")))

    left_edge = pl.concat([left_edge, left_edge_wraps]).with_columns(pl.lit("edge").alias("featureclass"))

    # Annotation on right edge
    right_edge = df.filter(pl.col("wraps") == False).join_where(features, (pl.col("left_bound") <= pl.col("end"))
                                            & (pl.col("end") <= pl.col("right_bound"))
                                            & (pl.col("start") < pl.col("left_bound")))
    right_edge_wraps = df.filter(pl.col("wraps") == True).join_where(features, (
                                            pl.col("left_bound") <= pl.col("end"))
                                            & (pl.col("right_bound") >= pl.col("end")))
    right_edge = pl.concat([right_edge, right_edge_wraps]).with_columns(pl.lit("edge").alias("featureclass"))

    # Annotation inside
    inside = df.filter(pl.col("wraps") == False).join_where(features, (pl.col("start") <= pl.col("left_bound")) & (pl.col("end") >= pl.col("right_bound")))
    inside_wraps_a = df.filter(pl.col("wraps") == True).join_where(features, (pl.col("start") <= pl.col("left_bound")))
    inside_wraps_b = df.filter(pl.col("wraps") == True).join_where(features, (pl.col("end") >= pl.col("right_bound")))
    inside = pl.concat([inside, inside_wraps_a, inside_wraps_b]).with_columns(pl.lit("inside").alias("featureclass"))

    # Annotation wraps
    wraps = df.filter(pl.col("wraps") == False).join_where(features, (pl.col("start") >= pl.col("left_bound")) & (pl.col("end") <= pl.col("right_bound")))
    wraps = pl.concat([wraps]).with_columns(pl.lit("wraps").alias("featureclass"))

    anno = pl.concat([left_edge, right_edge, inside, wraps]).unique()
    anno = anno.group_by(["repeat", "repeat_len", "first_repeat", "second_repeat", "distance", "mutation_rate"]
                        ).agg(pl.col('annotations'), pl.col('annotationobjects'))

    intergenic = df.join(anno.select(["repeat", "repeat_len", "first_repeat", "second_repeat", "distance"]), how="anti", on=["repeat", "repeat_len", "first_repeat", "second_repeat", "distance"]
    ).with_columns(pl.lit([]).alias("annotations")).with_columns(pl.lit([]).alias("annotationobjects")).select(["repeat", "repeat_len", "first_repeat", "second_repeat", "distance", "mutation_rate", "annotations", "annotationobjects"])

    df = pl.concat([anno, intergenic])

    return df


FASTA_EXTS = [".fa", ".fasta"]
GBK_EXTS = [".gb", ".gbk", ".gbff"]

def determine_looparound(df):
    """Determines if the prediction wraps around the origin"""
    pass

def sequence_to_features_df(sequence, circular=True):
    """Takes genbank annotations and turns them into a polars dataframe"""
    features = sequence.features
    seqlen = len(sequence)

    def get_feature_bounds(feature_location):
        if isinstance(feature_location, SimpleLocation):
            return feature_location.start, feature_location.end
        elif isinstance(feature_location, CompoundLocation):
            if feature_location.parts[-1].end < feature_location.parts[0].start:
                end = len(sequence) + feature_location.parts[-1].end
            else:
                end = feature_location.parts[-1].end
            return feature_location.parts[0].start, end

    if not circular:
        # Look for compound wraparounds and break them into simple features
        newfeatures = []
        deletedfeatures = []
        for feature in features:
            if not isinstance(feature.location, CompoundLocation):
                continue

            # Check whether the compound feature is actually a wraparound
            wraparound_part_index = None
            last_part_start = None
            rightmost_part = None
            for i, part in enumerate(feature.location.parts):
                if rightmost_part != None and part.start < last_part_start:
                    wraparound_part_index = i
                if rightmost_part == None:
                    rightmost_part = i
                    last_part_start = part.start
            if wraparound_part_index is None:
                continue

            # If it is a wraparound, break it into two features
            leftsplit_locations = feature.location.parts[:wraparound_part_index]
            if len(leftsplit_locations) > 1:
                leftsplit_locations = CompoundLocation(leftsplit_locations)
            else:
                leftsplit_locations = leftsplit_locations[0]
            newleftfeature = SeqFeature(leftsplit_locations, feature.type, qualifiers=feature.qualifiers)


            rightsplit_locations = feature.location.parts[wraparound_part_index:]
            if len(rightsplit_locations) > 1:
                rightsplit_locations = CompoundLocation(rightsplit_locations)
            else:
                rightsplit_locations = rightsplit_locations[0]
            newrightfeature = SeqFeature(rightsplit_locations, feature.type, qualifiers=feature.qualifiers)

            newfeatures.append(newleftfeature)
            newfeatures.append(newrightfeature)
            deletedfeatures.append(feature)

        for feature in deletedfeatures:
            features.remove(feature)
        features.extend(newfeatures)

    df = pl.DataFrame([(feature.type,
                        get_feature_bounds(feature.location),
                        feature.qualifiers.get("label", ["unlabeled"])[0],
                        seqfeature_hash(feature)) for feature in features],
        schema=['type', 'loc', 'annotations', 'annotationobjects'],
        orient="row")

    # expand out loc
    df = df.with_columns(pl.col("loc").list.to_struct(fields=['left_bound', 'right_bound'])).unnest("loc")
    df = df.with_columns(
        pl.col("left_bound").cast(pl.Int32),
        pl.col("right_bound").cast(pl.Int32)
    )

    return df



import streamlit as st
import subprocess
import io
import glob
import Bio.SeqIO as SeqIO
from Bio.SeqRecord import SeqRecord
from Bio.SeqFeature import CompoundLocation
from Bio.Seq import Seq
import base64
import zipfile
import hmac
import polars as pl
import os
from ..constants import VALID_EXTS, MAX_SIZE
from .bokeh_plot import bokeh_plot
from ..pipeline.primary_pipeline import predict
from ..ingest import parse_file, validate_sequences, BadSequenceError

from bokeh.embed import file_html
import streamlit.components.v1 as components
from tempfile import TemporaryDirectory
from pathlib import Path
from streamlit_extras.stylable_container import stylable_container
from streamlit_extras.add_vertical_space import add_vertical_space
from ..pipeline.mutation_rates import rip_score
from .vis_utils import eval_top
from ..StateMachine import StateMachine
from ..ingest import EFMSequence
from time import sleep

from streamlit_javascript import st_javascript

import hashlib

ASSET_LOCATION = os.path.join(os.path.dirname(__file__), "assets")

import pandas as pd
import base64
import json




def download_data(): # https://gist.github.com/snehankekre/2dcce9fb42b2f7e1742de7431326b263
    with TemporaryDirectory() as tempdir:
        outputdir = tempdir + "/results"
        os.mkdir(outputdir)
        statemachine = st.session_state["statemachine"]
        filetype = st.session_state["dlft"]
        statemachine.save_results(outputdir, filetype=filetype)
        filestream=io.BytesIO() # https://stackoverflow.com/questions/75304410/streamlit-download-button-not-working-when-trying-to-download-files-as-zip
        with zipfile.ZipFile(filestream, mode='w', compression=zipfile.ZIP_DEFLATED) as zipf:
            for root, dirs, files in os.walk(outputdir):
                    for file in files:
                        zipf.write(os.path.join(root, file),
                                    os.path.relpath(os.path.join(root, file),
                                                    os.path.join(outputdir, '..')))
        b64 = base64.b64encode(filestream.getvalue()).decode()

    dl_link = f"""
    <html>
    <head>
    <title>Start Auto Download file</title>
    <script src="http://code.jquery.com/jquery-3.2.1.min.js"></script>
    <script>
    $('<a href="data:application/zip;base64,{b64}" download="results.zip">')[0].click()
    </script>
    </head>
    </html>
    """
    components.html(
        dl_link,
        height=0,
    )


def check_feats_look_circular(seq):
    """Checks to see if features look circular."""
    features = seq.features
    for feature in features:
        if not isinstance(feature.location, CompoundLocation):
            continue
        # Check whether the compound feature is actually a wraparound
        wraparound_part_index = None
        rightmost_part = None
        last_part_start = None
        for i, part in enumerate(feature.location.parts):
            if rightmost_part != None and part.start < last_part_start:
                return True
            if rightmost_part == None:
                rightmost_part = i
                last_part_start = part.start
    else:
        return False

def run_webapp():

    st._config.set_option(f"theme.base", "light")
    st.set_page_config(
        page_title="EFM Calculator",
        page_icon=ASSET_LOCATION + "/favicon.ico",
        layout="wide",
        initial_sidebar_state="auto",
        menu_items=None,
    )

    st.markdown("""
    <style>
        .block-container {
            padding-top: 0rem;
            padding-bottom: 0rem;
            max-width: 1500px;
        }
    </style>
    """, unsafe_allow_html=True)

    st.markdown("""
    <style>
        #MainMenu, header, footer {visibility: hidden;}

        /* This code gets the first element on the sidebar,
        and overrides its default styling */
        section[data-testid="stSidebar"] div:first-child {
            top: 0;
            height: 100vh;
        }
    </style>
    """,unsafe_allow_html=True)

    st.markdown( # https://discuss.streamlit.io/t/image-and-text-next-to-each-other/7627/17
        """
        <style>
        .container {
            display: flex;
        }
        .logo-text {
            font-weight:700 !important;
            font-size:50px !important;
            color: #000000 !important;
            padding-top: 75px !important;
        }
        .logo-img {
            float:right;
        }
        </style>
        """,
        unsafe_allow_html=True
    )

    st.markdown("""
    <style>
        * {
           overflow-anchor: none !important;
           }
    </style>""", unsafe_allow_html=True)

    collogo,_,colbadge = st.columns([2,1,2], vertical_alignment="bottom")
    with collogo:
        st.markdown(
            f"""
            <div class="container">
                <img class="logo-img" src="data:image/svg+xml;base64,{base64.b64encode(open(ASSET_LOCATION + "/tombstone.svg", "rb").read()).decode()}">
                <p class="logo-text">EFM Calculator</p>
            </div>
            """,
            unsafe_allow_html=True
        )
        try:
            from .._version import version_tuple
            st.markdown(f"Version {version_tuple[0]}.{version_tuple[1]}.{version_tuple[2]} ([{str(version_tuple[4])[1:8]}](https://www.github.com/barricklab/efm-calculator2/commit/{str(version_tuple[4]).split('.')[0][1:8]}))")
        except:
            pass
    with colbadge:
        st.html(r'<a href="https://github.com/barricklab/efm-calculator2"><img alt="GitHub Repo stars" src="https://img.shields.io/github/stars/barricklab/efm-calculator2?style=social&label=barricklab%2Fefm-calculator2"></a>')


    col1,col2,col3 = st.columns([2,1,2])

    with col1:
        upload_option = "Upload files (FASTA, GenBank, or CSV)"
        enter_option = "Copy/Paste Plain Text"
        example_option = "Example"

        option = st.radio(
            "Choose method of submitting sequence:",
            [example_option, upload_option, enter_option],
        )

        inSeq = None

    with col3:
        st.write("The EFM Calculator predicts mutational hotspots as a result of DNA polymerase slippage. It classifies these hotspots into three categories, Short Sequence Repeats, Short Repeated Sequences, and Repeat Mediated Deletions. For more information, please see the paper. If you have found this tool helpful, please remember to cite it as well.")
        st.write("Jack, B. R., Leonard, S. P., Mishler, D. M., Renda, B. A., Leon, D., Suárez, G. A., & Barrick, J. E. (2015). Predicting the Genetic Stability of Engineered DNA Sequences with the EFM Calculator. ACS Synthetic Biology, 4(8), 939–943. https://doi.org/10.1021/acssynbio.5b00068")


    # Initialize session state
    if not st.session_state.get("statemachine", False):
        st.session_state["statemachine"] = StateMachine()
    statemachine = st.session_state["statemachine"]

    with TemporaryDirectory() as tempdir:
        is_circular = True
        if option == upload_option:
            with col1:
                is_circular = st.checkbox(label="Circular Prediction", value=True)
            upload_disclaimer = f"Total sequence length must be less than {MAX_SIZE+1}. CSV files must have a 'seq' column and may have a 'name' column."
            uploaded_files = st.file_uploader("Choose a file:", type=VALID_EXTS, accept_multiple_files = True)
            st.write(upload_disclaimer)
            if uploaded_files:
                inSeq = []
                for uploaded_file in uploaded_files:
                    original_filename = os.path.splitext(uploaded_file.name)[0]
                    uploaded_filetype = Path(uploaded_file.name).suffix
                    # Hash the file to create a safe name
                    filename = Path(
                        tempdir + str(hash(uploaded_file.name)) + uploaded_filetype
                    )
                    with open(filename, "wb") as f:
                        f.write(uploaded_file.getbuffer())
                    sequences = parse_file(filename, use_filename=False, iscircular = is_circular)
                    file_sequences = []

                    for sequence in sequences:
                        filename = original_filename
                        if not sequence.name:
                            sequence.name = f"{filename}"
                        if not sequence.description or sequence.description == '':
                            sequence.description = f"{filename}"
                        file_sequences.append(sequence)

                    inSeq.extend(file_sequences)

                st.success("Files uploaded.")

        elif option == enter_option:
            with col1:
                is_circular = st.checkbox(label="Circular Prediction", value=True)
            upload_disclaimer = f"""<div>
            <p>Total sequence length must be less than {MAX_SIZE+1}.</p>
            </div>"""
            field = st.text_area("Input sequence here:", max_chars=MAX_SIZE)
            st.markdown(upload_disclaimer, unsafe_allow_html=True)
            field = field.replace("\n", "")
            field = field.replace(" ", "")
            field = "".join([i for i in field if not i.isdigit()])
            if field:
                record = SeqRecord(Seq(field), id="sequence")
                originhash = hashlib.md5(("string" + field).encode())
                record = EFMSequence(record, is_circular, originhash)
                inSeq = [record]

        elif option == example_option:
            with col1:
                gbs = []
                examples_path = "examples/"
                for infile_loc in glob.glob(os.path.join(examples_path, "*.gb")) + glob.glob(os.path.join(examples_path, "*.fasta")):
                    gbs.append(infile_loc.split("/")[-1])
                gbs = sorted(gbs)
                exampleFile = st.radio("Choose example file:", gbs)
                filepath = Path(examples_path + f"{exampleFile}")
                if filepath:
                    inSeq = parse_file(filepath, iscircular = True)
                st.write("Examples run in linear mode")


        if not inSeq:
            st.stop()

        statemachine.import_sequences(inSeq, max_size=50000, webapp = True)

        if len(inSeq) == 1:
            disable_dropdown = True
        else:
            disable_dropdown = False

        col4,col5,col6 = st.columns([2,1,2])

        with col4:
            selected_sequence = st.selectbox(
                "Sample:", statemachine.named_sequences.keys(),
                disabled = disable_dropdown
            )
            selectedhash = statemachine.named_sequences[selected_sequence]
            seq_record = statemachine.sequencestates[selectedhash]

        with col6:
                with TemporaryDirectory() as tempdir:
                    submit = st.button("Download results",
                                       on_click=download_data,
                                       use_container_width=True,
                                       type="primary")
                st.selectbox("Download File Format",["csv", "parquet"], key="dlft",
                    help="CSV files are easily usable in most spreadsheat programs but lack annotation information. Parquet files require specialized tooling but include annotation metadata")

        unique_features = seq_record.unique_annotations

        if not seq_record.predicted:
            seq_record.efmsequence.call_predictions(strategy="pairwise")
            seq_record.post_predict_processing()

        figcontainer = st.container(height=340)

        if unique_features:
            feature_filter = st.multiselect('Filter by feature annotation',
                                            sorted(unique_features),
                                            default=seq_record._last_filters)
        else:
            feature_filter = []
        seq_record.set_filters(feature_filter)

        if not st.session_state.get("last_filter", []) == feature_filter:
            seq_record.reset_selected_predictions()
        st.session_state["last_filter"] = feature_filter

        results = [seq_record._filtered_ssrs, seq_record._filtered_srss, seq_record._filtered_rmds]

        if feature_filter:
            sequence_of_interest = seq_record.annotation_coverage(feature_filter)
        else:
            sequence_of_interest  = len(seq_record.seq)

        summary = rip_score(results[0], results[1], results[2], sequence_length = sequence_of_interest)
        looks_circular = check_feats_look_circular(seq_record)
        if looks_circular:
            st.warning("You deselected the circular option, but your file looks circular.", icon="⚠️")
        col7, col8 = st.columns(2)
        with col7:
            st.markdown(f"<div style='text-align: center;'>RIP score: {summary['rip']:.2f}</div>", unsafe_allow_html=True)
        with col8:
            if summary['ssr_sum'] > 0:
                st.markdown(f"<div style='text-align: center;'>SSRs: {summary['ssr_sum']:.2e}</div>", unsafe_allow_html=True)
            else:
                st.markdown(f"<div style='text-align: center;'>SSRs: 0</div>", unsafe_allow_html=True)
            if summary['srs_sum'] > 0:
                st.markdown(f"<div style='text-align: center;'>SRSs: {summary['srs_sum']:.2e}</div>", unsafe_allow_html=True)
            else:
                st.markdown(f"<div style='text-align: center;'>SRSs: 0</div>", unsafe_allow_html=True)
            if summary['rmd_sum'] > 0:
                st.markdown(f"<div style='text-align: center;'>RMDs: {summary['rmd_sum']:.2e}</div>", unsafe_allow_html=True)
            else:
                st.markdown(f"<div style='text-align: center;'>RMDs: 0</div>", unsafe_allow_html=True)
            if summary['bps_sum'] > 0:
                st.markdown(f"<div style='text-align: center;'>Base Pair Substitution Rate: {summary['bps_sum']:.2e}</div>", unsafe_allow_html=True)
            else:
                st.markdown(f"<div style='text-align: center;'>Base Pair Substitution Rate: 0</div>", unsafe_allow_html=True)

        tab1, tab2, tab3, tab4 = st.tabs(["Top", "SSR", "SRS", "RMD"])
        seq_record.refresh_last_shown()
        seq_record.rebuild_top_table()
        seq_record.rebuild_ssr_table()
        seq_record.rebuild_srs_table()
        seq_record.rebuild_rmd_table()

        with tab1:
            seq_record.top_webapp_table

        with tab2:
            seq_record.ssr_webapp_table

        with tab3:
            seq_record.srs_webapp_table

        with tab4:
            seq_record.rmd_webapp_table

        with figcontainer:
            fig = bokeh_plot(seq_record)
            sleep(0.1) # Helps to avoid two plots displayed
            st.bokeh_chart(fig, use_container_width=True)

    add_vertical_space(4)
    seq_record.refreshed = False



import pandas as pd
from sklearn.model_selection import train_test_split
from pygam import LinearGAM, s

#load csv file
data_of_interest = pd.read_csv('test.csv')
print(data_of_interest.columns)


#separating df
X = data_of_interest[['RBP_Length', 'TBD_length']]
y = data_of_interest['Mutation_Rate']
weights = data_of_interest['weight']

#gam model fit with cr splines and regularization b/c small data
gam_model = LinearGAM(s(0, n_splines=3, spline_order=2) +  s(1, n_splines=3, spline_order=2)) 
gam_model.fit(X, y, weights=weights
              )

#data not included
new_data = pd.DataFrame({'RBP_Length': [6], 'TBD_length': [50]})


#prediction site
predictions = gam_model.predict(new_data)
predictions = (predictions)

print(gam_model.summary())
print("Mutation Rate:", predictions)





import argparse
import logging
import pathlib
import time
from importlib.metadata import version, PackageNotFoundError
import os

from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from .utilities import (
    is_path_creatable,
    is_pathname_valid,
)
from .ingest.EFMSequence import EFMSequence
from .ingest.bad_state_mitigation import BadSequenceError
from .ingest.parse_inputs import (
    parse_file,
    validate_sequence,
    validate_sequences,
)

from .constants import VALID_STRATEGIES, VALID_FILETYPES
from .StateMachine import StateMachine

import logging
from rich.logging import RichHandler

FORMAT = "%(message)s"
logging.basicConfig(
    level="NOTSET", format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
)
logger = logging.getLogger(__name__)
logging.getLogger(__name__).addHandler(logging.NullHandler())

def main():
    """CLI entry point for EFM Calculator"""

    start_time = time.time()

    # Parse args

    parser = argparse.ArgumentParser(description="Find Mutation Hotspots from Input Sequences")
    parser.add_argument(
        "-i",
        "--input",
        action="store",
        dest="inpath",
        required=True,
        help="the path to fasta/genbank/csv file",
    )
    parser.add_argument(
        "-o",
        "--output",
        action="store",
        dest="outpath",
        required=True,
        help="path to output prefix",
    )
    parser.add_argument(
        "-s",
        "--strategy",
        action="store",
        dest="strategy",
        required=False,
        default="pairwise",
        help="the strategy to use for predicting deletions. Must be one of 'pairwise' or 'linear' (default: 'pairwise')",
    )
    parser.add_argument(
        "-c",
        "--circular",
        dest="circular",
        action="store_true",
        required=False,
        help="Is circular?",
    )
    parser.add_argument(
        "-f",
        "--filetype",
        dest="filetype",
        action="store",
        required=False,
        help="Output filetype for tables (csv | parquet)",
    )
    parser.add_argument(
        "-j",
        "--maxthreads",
        dest="threads",
        action="store",
        type=int,
        required=False,
        help="Maximum number of threads (>0)",
    )
    parser.add_argument(
        "-t",
        "--tall",
        dest="tall",
        action="store_true",
        required=False,
        help="Parallelize across samples rather than within samples",
    )
    parser.add_argument(
        "-v",
        "--verbosity",
        action="store",
        type=int,
        dest="verbose",
        required=False,
        default=1,
        help="0 - Silent | 1 Basic Information | 2 Debug",
    )
    parser.add_argument(
        "--summary",
        dest="summaryonly",
        action="store_true",
        required=False,
        help="Only save summary information rather than all . Useful for very tall inputs.",
    )

    try:
        from ._version import version_tuple
        pkgversion = f"{version_tuple[0]}.{version_tuple[1]}.{version_tuple[2]}"
    except:
        pkgversion = "unknown"

    logger.info("EFM Calculator {}".format(pkgversion))

    args = parser.parse_args()

    # Set up logger  -----------
    logging.basicConfig()
    if args.verbose == 0:
        logging.root.setLevel(logging.ERROR)
    elif args.verbose == 1:
        logging.root.setLevel(logging.INFO)
    elif args.verbose == 2:
        logging.root.setLevel(logging.DEBUG)
    else:
        logger.error(f"Invalid value for verbosity flag '{args.verbose}'")
        parser.print_help()
        exit(1)

    # Sanity checks  ------------
    if args.strategy not in VALID_STRATEGIES:
        logger.error(f"Invalid value for strategy flag '{args.strategy}'")
        parser.print_help()
        exit(1)

    if args.filetype and args.filetype not in VALID_FILETYPES:
        logger.error(f"Invalid value for filetype flag '{args.filetype}'")
        parser.print_help()
        exit(1)
    else:
        args.filetype = "csv"

    if not is_pathname_valid(args.inpath):
        logger.error(f"File {args.inpath} is not a valid path.")
        exit(1)
    elif not is_pathname_valid(args.outpath):
        logger.error(f"File {args.outpath} is not a valid path.")
        exit(1)
    elif not is_path_creatable(args.outpath):
        logger.error(f"Cannot write to {args.outpath}")
        exit(1)


    if args.tall:
        os.environ["POLARS_MAX_THREADS"] = "1"
    if args.threads is not None and args.threads <=0:
        logger.error("Max threads must be greater than 0")
        exit(1)
    elif args.tall and not args.threads:
        threads = os.cpu_count()
    else:
        threads = args.threads
        os.environ["POLARS_MAX_THREADS"] = str(threads)

    global pl
    import polars as pl


    # Set up circular ------------

    args.isCirc = args.circular

    # Grab sequence information --------
    try:
        sequences = parse_file(pathlib.Path(args.inpath), iscircular=args.circular)
    except ValueError as e:
        logger.error(e)
        exit(1)
    except OSError as e:
        try:
            validate_sequence(EFMSequence(SeqRecord(Seq(args.inpath), id="text input", name="text input"), is_circular=args.circular))
            sequences = [EFMSequence(SeqRecord(Seq(args.inpath), id="text input", name="text input"), is_circular=args.circular)]
        except:
            logger.error("Input is not an existing file or valid sequence")
            exit(1)


    # Unpack sequences into list ---------
    sequences = list(sequences)

    # Run EFM Calculator ----------------
    statemachine = StateMachine()

    try:
        statemachine.import_sequences(sequences)
    except BadSequenceError as e:
        logger.error(e)
        exit(1)

    if args.tall:
        statemachine.predict_tall(strategy=args.strategy,
                                  outpath=args.outpath,
                                  filetype=args.filetype,
                                  threads=threads,
                                  keepmem=not args.summaryonly,
                                  summaryonly=args.summaryonly)
    else:
        for i, seqobject in enumerate(statemachine.user_sequences.values()):
            logger.info(msg=f"Running on sequence {i}: {str(seqobject)}")
            seqobject.call_predictions(strategy=args.strategy)
        statemachine.save_results(args.outpath, filetype=args.filetype, summaryonly=args.summaryonly)


    # Done ------------------------------

    # Logging
    t = time.time() - start_time
    t_sec = int(t)
    t_msec = round((t - t_sec) * 1000)
    t_min, t_sec = divmod(t_sec, 60)
    t_hour, t_min = divmod(t_min, 60)
    logger.info(
        f"EFMCalculator completed in {t_hour:02d}h:{t_min:02d}m:{t_sec:02d}s:{t_msec:02d}ms"
    )

if __name__ == "__main__":
    main()



import polars as pl
import os
import sys
from ..constants import SUB_RATE

try:
    gam_result_filepath = filepath = os.path.dirname(sys.modules['efmcalculator'].__file__)+'/data/gam_df.csv'
except e:
    print("Error while loading gam results")
    exit(1)
gam_results = pl.read_csv(gam_result_filepath).cast({"repeat_len": pl.Int32,
                                            "distance": pl.Int32})


def ssr_mut_rate(repeat_count, unit_length, org):
    """
    Calculates mutation rate for simple sequence repeats
    :param repeat_count: Number of times the repeating unit occurs
    :param unit_length: Length of repeating unit
    :param org: Host organism
    :return: Mutation rate
    """
    mut_rate = float(0)
    if org == "ecoli" or org == "reca":
        if unit_length == 1:
            # Formula based on analysis of Lee et. al. data
            mut_rate = float(10 ** (0.72896 * repeat_count - 12.91471))
        elif unit_length > 1:
            mut_rate = float(10 ** (0.06282 * repeat_count - 4.74882))
    elif org == "yeast":
        if unit_length == 1:
            mut_rate = float(10 ** (0.3092 * repeat_count - 7.3220))
        elif unit_length > 1:
            mut_rate = float(10 ** (0.11141 * repeat_count - 7.65810))
    return mut_rate


def ssr_mut_rate_vector(ssr_df, org="ecoli"):
    if org == "ecoli" or org == "reca":
        ssr_df = (
            ssr_df.lazy().with_columns(
                mutation_rate=(
                    pl.when(pl.col("repeat_len") == 1)
                    .then(10 ** (0.72896 * pl.col("count") - 12.91471))
                    .otherwise(10 ** (0.06282 * pl.col("count") - 4.74882))
                )
            )
        ).collect()

    elif org == "yeast":
        ssr_df = (
            ssr_df.lazy().with_columns(
                mutation_rate=(
                    pl.when(pl.col("repeat_len") == 1)
                    .then(10 ** (0.3092 * pl.col("count") - 7.3220))
                    .otherwise(10 ** (0.11141 * pl.col("count") - 7.65810))
                )
            )
        ).collect()
    return ssr_df


def rmd_mut_rate(length, distance, org):
    """
    Calculate the recombination rate based on the Oliviera, et. al. formula
    :param length: Length of homologous region
    :param distance: Bases between the end of the first repeat and beginning of second
    :param org: Host organism
    :return: Recombination rate
    """
    spacer = distance
    # If the homologous sequences overlap we can't calculate a rate
    if spacer < 0:
        return 0
    if org == "ecoli" or org == "yeast":
        recombo_rate = float(
            ((8.8 + spacer) ** (-29.0 / length)) * (length / (1 + 1465.6 * length))
        )
    elif org == "reca":
        recombo_rate = float(
            ((200.4 + spacer) ** (-8.8 / length))
            * (length / (1 + 2163.0 * length + 14438.6 * spacer))
        )
    else:
        raise ValueError("Invalid org")

    return recombo_rate


def rmd_mut_rate_vector(rmd_df, org="reca"):
    if org == "ecoli" or org == "yeast":
        rmd_df = (
            rmd_df.lazy().with_columns(
                mutation_rate=(
                    ((8.8 + pl.col("distance")) ** (-29.0 / pl.col("repeat_len")))
                    * (pl.col("repeat_len") / (1 + 1465.6 * pl.col("repeat_len")))
                )
            )
        ).collect()

    elif org == "reca":
        rmd_df = (
            rmd_df.lazy().with_columns(
                mutation_rate=(
                    (200.4 + pl.col("distance")) ** (-8.8 / pl.col("repeat_len"))
                )
                * (
                    pl.col("repeat_len")
                    / (1 + 2163.0 * pl.col("repeat_len") + 14438.6 * pl.col("distance"))
                )
            )
        ).collect()
    return rmd_df


def srs_mut_rate_vector(rmd_df, org="reca"):
    """
    Calculate the recombination rate based on a GAM model.

    :param length: Length of the homologous region (RBP_length).
    :param distance: Bases between the end of the first repeat and the beginning of the second (TBD_length).
    :param SRS_df: DataFrame containing the GAM model predictions with columns 'RBP_Length', 'TBD_length', and 'prediction'.
    :return: The predicted mutation rate or 0 if the input is invalid.
    """


    findings = rmd_df.join(gam_results, on=['distance', 'repeat_len'], how='left').with_columns(pl.col("mutation_rate").fill_null(strategy="zero"))
    return findings


def rip_score(ssr_df, srs_df, rmd_df, sequence_length):
    if isinstance(ssr_df, pl.DataFrame) and not ssr_df.is_empty():
        ssr_sum = ssr_df.select(pl.sum("mutation_rate")).item()
    else:
        ssr_sum = 0
    if isinstance(srs_df, pl.DataFrame) and not srs_df.is_empty():
        srs_sum = srs_df.select(pl.sum("mutation_rate")).item()
    else:
        srs_sum = 0
    if isinstance(rmd_df, pl.DataFrame) and not rmd_df.is_empty():
        rmd_sum = rmd_df.select(pl.sum("mutation_rate")).item()
    else:
        rmd_sum = 0

    base_rate = float(sequence_length) * float(SUB_RATE)
    # Add in the mutation rate of an individual nucleotide
    r_sum = float(ssr_sum + srs_sum + rmd_sum  + base_rate)

    # Set the maximum rate sum to 1 for now.
    if r_sum > 1:
        r_sum = float(1)
    rel_rate = float(r_sum) / float(base_rate)

    return {
        "rip": rel_rate,
        "ssr_sum": ssr_sum,
        "srs_sum": srs_sum,
        "rmd_sum": rmd_sum,
        "bps_sum": base_rate,
    }



def detect_special_cases(sequence):
    def can_tile(s):
        doubled_s = s + s
        modified_doubled = doubled_s[1:-1]
        return s in modified_doubled
    if can_tile(str(sequence).lower()) and sequence.is_circular:
        raise BadSequenceError("Circular sequence is an infinitely long SSR. Did you mean to use a linear strategy?")
    if len(str(sequence).lower()) < 21 and sequence.is_circular:
        raise BadSequenceError("EFM Calculator is limited to circular sequences of at least 21 bases.")


class BadSequenceError(ValueError):
    pass



from subprocess import run
import bokeh as bk
import copy
from Bio import SeqFeature
from bokeh.plotting import figure
from bokeh.models import HoverTool, CustomJS
from typing import List
import polars as pl
from ..constants import MARKER_HEIGHT, COLORS
from copy import deepcopy

OUTLINE_PADDING_X = 25
OUTLINE_PADDING_Y = 25
INNER_HEIGHT = MARKER_HEIGHT-OUTLINE_PADDING_Y*2
ANNOTATION_HEIGHT = MARKER_HEIGHT/2

def bokeh_plot(seqobj):
    # Figure boilerplate

    fig = figure(plot_height=300, output_backend="svg")
    fig.line(x=[0, len(seqobj.seq)], y=[0, 0], line_color="black", line_width=2)
    fig.xaxis.axis_label = "Position"
    fig.yaxis.visible = False
    fig.ygrid.visible = False
    fig.xgrid.visible = False
    fig.toolbar.logo = None
    fig, depth = plot_annotations(fig, seqobj)

    # Calculate vertical stagger to prevent overlapping annotations
    selected_ssrs = seqobj.get_shown_predictions(seqobj._filtered_ssrs)
    selected_ssrs = seqobj._filtered_ssrs.with_row_index(
                           ).filter(pl.col("index").is_in(selected_ssrs))
    selected_srss = seqobj.get_shown_predictions(seqobj._filtered_srss)
    selected_srss = seqobj._filtered_srss.with_row_index(
                           ).filter(pl.col("index").is_in(selected_srss))
    selected_rmds = seqobj.get_shown_predictions(seqobj._filtered_rmds)
    selected_rmds = seqobj._filtered_rmds.with_row_index(
                           ).filter(pl.col("index").is_in(selected_rmds))

    stagger_database = {
        0: []
    }

    stagger_ssrs = selected_ssrs.with_columns(
        (pl.col("repeat_len")*pl.col("count")+pl.col("start")-1).alias("end").cast(pl.Int32),
        pl.lit("SSR").alias("type"),
        pl.col("start").alias("first_repeat"),
        pl.lit(None).cast(pl.Int32).alias("second_repeat"),
    )
    stagger_srss = selected_srss.with_columns(
        (pl.col("second_repeat") + pl.col("repeat_len")).alias("end").cast(pl.Int32),
        pl.col("first_repeat").alias("start"),
        pl.lit("SRS").alias("type"),
    )
    stagger_rmds = selected_rmds.with_columns(
        (pl.col("second_repeat") + pl.col("repeat_len")).alias("end").cast(pl.Int32),
        pl.col("first_repeat").alias("start"),
        pl.lit("RMD").alias("type"),
    )

    joint_table = pl.concat([stagger_ssrs,
                             stagger_srss,
                             stagger_rmds],
                             how="diagonal").with_columns(
        (pl.col('end') - pl.col('start') + pl.lit(1)).alias("size")
                             ).select(["start", "end", "size", "predid", "type"]).with_columns(
        pl.lit(0).alias("level")
                             )
    joint_table = joint_table.sort(by=["size"], descending=True)

    highest_level = 0
    for i, row in enumerate(joint_table.rows(named=True)):
        assigned = False
        for level in stagger_database:
            for occupied_area in stagger_database[level]:
                if occupied_area[0] <= row['start'] and occupied_area[1] >= row['end']:
                    # Entirely inside
                    break
                elif occupied_area[0] <= row['start'] and row['start'] <= occupied_area[1] <= row['end']:
                    # Left edge
                    break
                elif  row['start'] <= occupied_area[0] <= row['end'] and occupied_area[1] >= row['end']:
                    # Right edge
                    break
            else:
                # Outside all areas
                stagger_database[level].append([row['start']-OUTLINE_PADDING_X*2, row['end']+OUTLINE_PADDING_X*2])
                joint_table[i, 'level'] = level
                assigned = True
                break
        else:
            # All existing levels are full
            highest_level += 1
            stagger_database[highest_level] = [[row['start']-OUTLINE_PADDING_X*2, row['end']+OUTLINE_PADDING_X*2]]
            joint_table[i, 'level'] = highest_level
            assigned = True

    joint_table = joint_table.select(["predid", "level"])
    stagger_ssrs = stagger_ssrs.join(joint_table, on=["predid"], how="left")
    stagger_srss = stagger_srss.join(joint_table, on=["predid"], how="left")
    stagger_rmds = stagger_rmds.join(joint_table, on=["predid"], how="left")

    fig = plot_rmd(fig, stagger_rmds)
    fig = plot_srs(fig, stagger_srss)
    fig = plot_ssr(fig, stagger_ssrs)


    return fig

def plot_annotations(fig, seqobj):

    def assign_feature_levels(features: List[SeqFeature]) -> List[SeqFeature]:
        end_positions = dict()
        for feature in features:
            if not end_positions:
                end_positions[str(0)] = feature.location.end
                feature.qualifiers["nest_level"] = 0
                continue
            for level, position in end_positions.items():
                level = int(level)
                if position == "inf":
                    continue
                elif int(position) > int(feature.location.start):
                    continue
                else:
                    end_positions[str(level)] = feature.location.end
                    feature.qualifiers["nest_level"] = level
                    break
            if (
                not feature.qualifiers.get("nest_level")
                and feature.qualifiers.get("nest_level") != 0
            ):  # Layer 0 is falcey
                lowest_level = max([int(n) for n in end_positions])
                end_positions[str(lowest_level + 1)] = feature.location.end
                feature.qualifiers["nest_level"] = lowest_level + 1
        return features

    def get_feature_color(feature: SeqFeature) -> str:
        feature_type = feature.type
        if feature.qualifiers.get("ApEinfo_fwdcolor"):
            return feature.qualifiers.get("ApEinfo_fwdcolor")[0]
        elif feature.qualifiers.get("ApEinfo_revcolor"):
            return feature.qualifiers.get("ApEinfo_revcolor")[0]
        elif feature_type in ["gene"]:
            color = COLORS["cds"]
        elif feature_type in ["promoter"]:
            color = COLORS["promoter"]
        elif feature_type in ["terminator"]:
            color = COLORS["terminator"]
        elif feature_type in ["ori", "rep_origin"]:
            color = COLORS["ori"]
        elif feature_type in ["ncRNA"]:
            color = COLORS["ncrna"]
        elif feature_type in ["primer_bind"]:
            color = COLORS["primer_bind"]
        else:
            color = COLORS["misc"]
        return color

    staggered_features = assign_feature_levels(seqobj.features)
    xmax = len(seqobj.seq)

    annotation_depth = -500
    lowest_annotation_y = -ANNOTATION_HEIGHT
    genbank_dictionary = {
            "x": [],
            "y": [],
            "color": [],
            "name": [],
            "position": [],
            "strand": [],
        }
    for genbank_annotation in staggered_features:
        arrow_depth = 100
        # @TODO: SCALE ANNOTATIONS

        scaled_start = genbank_annotation.location.start
        scaled_end = genbank_annotation.location.end
        featurename = genbank_annotation.qualifiers.get("label")

        if (
            genbank_annotation.location.end - genbank_annotation.location.start
            < arrow_depth
        ):
            arrow_depth = (
                genbank_annotation.location.end - genbank_annotation.location.start
            )
        annotation_base_y = (
            -650.0 * genbank_annotation.qualifiers.get("nest_level", 0)
            + annotation_depth
        )

        if genbank_annotation.location.strand == 1:
            xs = [
                scaled_start,
                scaled_start,
                scaled_end - arrow_depth,
                scaled_end,
                scaled_end - arrow_depth,
            ]
            ys = [
                annotation_base_y + ANNOTATION_HEIGHT,
                annotation_base_y - ANNOTATION_HEIGHT,
                annotation_base_y - ANNOTATION_HEIGHT,
                annotation_base_y,
                annotation_base_y + ANNOTATION_HEIGHT,
            ]
        elif genbank_annotation.location.strand == -1:
            xs = [
                scaled_end,
                scaled_end,
                scaled_start + arrow_depth,
                scaled_start,
                scaled_start + arrow_depth,
            ]
            ys = [
                annotation_base_y - ANNOTATION_HEIGHT,
                annotation_base_y + ANNOTATION_HEIGHT,
                annotation_base_y + ANNOTATION_HEIGHT,
                annotation_base_y,
                annotation_base_y - ANNOTATION_HEIGHT,
            ]
        else:
            xs = [scaled_start, scaled_start, scaled_end, scaled_end]
            ys = [
                annotation_base_y - ANNOTATION_HEIGHT,
                annotation_base_y + ANNOTATION_HEIGHT,
                annotation_base_y + ANNOTATION_HEIGHT,
                annotation_base_y - ANNOTATION_HEIGHT,
            ]

        # Update lowest annotation y
        if annotation_base_y - ANNOTATION_HEIGHT < lowest_annotation_y:
            lowest_annotation_y = annotation_base_y - ANNOTATION_HEIGHT

        # Define the color
        color = get_feature_color(genbank_annotation)

        # Add the annotation to the dictionary

        genbank_dictionary["x"].append(xs)
        genbank_dictionary["y"].append(ys)
        genbank_dictionary["color"].append(color)
        genbank_dictionary["name"].append(featurename)
        genbank_dictionary["position"].append(
            f"{genbank_annotation.location.start}-{genbank_annotation.location.end}"
        )
        genbank_dictionary["strand"].append(genbank_annotation.location.strand)


    genbank_glyphs = fig.patches(
        "x",
        "y",
        color="color",
        source=genbank_dictionary,
        alpha=0.5,
        line_color="black",
        line_width=2,
    )
    genbank_glyphs_hover = HoverTool(
        renderers=[genbank_glyphs], tooltips=[("Name", "@name")]
    )
    fig.add_tools(genbank_glyphs_hover)

    # Draw a line below the annotations

    lowest_annotation_y -= ANNOTATION_HEIGHT/2
    fig.line(
        [0, xmax],
        [lowest_annotation_y, lowest_annotation_y],
        line_width=2,
        color="black",
    )

    return fig, lowest_annotation_y

def plot_ssr(fig, ssr_df):
    height = 1
    ssr_shape = [[0, 0], [1, 0], [1, INNER_HEIGHT], [0, INNER_HEIGHT], [0, 0]]
    color = "green"

    ssr_source = {
        "x": [],
        "y": [],
        "color": [],
        "line_color": [],
        "name": [],
        "first_repeat": [],
        "mutation_rate": [],
        "line_width": [],
        "sequence": [],
    }

    outline_shape_left = [[0, 0], [1, 0],  [1, MARKER_HEIGHT], [0, MARKER_HEIGHT], [0, 0]]
    ssr_outline_source = {
        "x": [],
        "y": [],
        "color": [],
        "line_color": [],
        "name": [],
        "first_repeat": [],
        "mutation_rate": [],
        "line_width": [],
        "sequence": [],
        "count": [],
        "monomer": [],
    }

    for row in ssr_df.rows(named=True):
        left_pos = row["start"]
        repeat_len = row["repeat_len"]
        y_height = row["level"]*MARKER_HEIGHT*1.5
        glyph = [[point[0]*repeat_len+left_pos, point[1]+OUTLINE_PADDING_Y] for point in ssr_shape]

        ssr_source["x"].append([x for x, _ in glyph])
        ssr_source["y"].append([y+y_height+MARKER_HEIGHT*0.5 for _, y in glyph])
        ssr_source["color"].append("black")
        ssr_source["name"].append("ssr")
        ssr_source['first_repeat'].append(left_pos)
        ssr_source["mutation_rate"].append(row["mutation_rate"])
        ssr_source["line_width"].append(2)
        ssr_source["line_color"].append("black")
        ssr_source["sequence"].append(row["repeat"]*repeat_len)

        # Outline
        outline = [[point[0]*repeat_len+left_pos, point[1]+OUTLINE_PADDING_Y] for point in ssr_shape]

        outline_mods = [[-OUTLINE_PADDING_X, -OUTLINE_PADDING_Y],
                        [OUTLINE_PADDING_X+repeat_len, -OUTLINE_PADDING_Y],
                        [OUTLINE_PADDING_X+repeat_len, OUTLINE_PADDING_Y],
                        [-OUTLINE_PADDING_X, OUTLINE_PADDING_Y],
                        [-OUTLINE_PADDING_X, -OUTLINE_PADDING_Y]]
        for i in range(len(outline)):
            outline[i][0] = outline[i][0] + outline_mods[i][0]
            outline[i][1] = outline[i][1] + outline_mods[i][1] + y_height + MARKER_HEIGHT*0.5
        ssr_outline_source["x"].append([x for x, _ in outline])
        ssr_outline_source["y"].append([y for _, y in outline])
        ssr_outline_source["color"].append(None)
        ssr_outline_source["name"].append("SSR")
        ssr_outline_source['first_repeat'].append(left_pos)
        ssr_outline_source["mutation_rate"].append(row["mutation_rate"])
        ssr_outline_source["line_width"].append(3)
        ssr_outline_source["line_color"].append(color)
        ssr_outline_source["sequence"].append(row["repeat"]*row["count"])
        ssr_outline_source["monomer"].append(row["repeat"])
        ssr_outline_source['count'].append(row["count"])

    ssr_glyphs = fig.patches(
        "x",
        "y",
        color="color",
        source=ssr_source,
        alpha=0.5,
        line_color="line_color",
        line_width="line_width",
    )

    ssr_glyphs_outline = fig.patches(
        "x",
        "y",
        color="color",
        source=ssr_outline_source,
        alpha=0.5,
        line_color="line_color",
        line_width="line_width",
    )

    ssr_glyphs_hover = HoverTool(
        renderers=[ssr_glyphs_outline],
        tooltips=[
            ("Type", "@name"),
            ("Sequence", "@sequence"),
            ("Repeat Start", "@first_repeat"),
            ("Monomer", "@monomer"),
            ("Num. Repeats", "@count"),
            ("Mutation Rate", "@mutation_rate"),
        ],
    )

    fig.add_tools(ssr_glyphs_hover)

    return fig

def plot_srs(fig, ssr_df):
    height = 1
    rmd_shape = [[0, 0], [1, 0], [1, INNER_HEIGHT], [0, INNER_HEIGHT], [0, 0]]
    color = "orange"

    # Logic for exact position
    srs_source = {
        "x": [],
        "y": [],
        "color": [],
        "line_color": [],
        "name": [],
        "first_repeat": [],
        "second_repeat": [],
        "mutation_rate": [],
        "line_width": [],
        "sequence": [],
        "distance": [],
    }

    for row in ssr_df.rows(named=True):
        left_pos = row["first_repeat"]
        right_pos = row["second_repeat"]
        repeat_len = row["repeat_len"]
        y_height = row["level"]*MARKER_HEIGHT*1.5
        left_glyph = [[point[0]*repeat_len+left_pos, point[1]+OUTLINE_PADDING_Y] for point in rmd_shape]
        right_glyph = [[point[0]*repeat_len+right_pos, point[1]+OUTLINE_PADDING_Y] for point in rmd_shape]

        srs_source["x"].append([x for x, _ in left_glyph])
        srs_source["y"].append([y+y_height+MARKER_HEIGHT*0.5 for _, y in left_glyph])
        srs_source["x"].append([x for x, _ in right_glyph])
        srs_source["y"].append([y+y_height+ MARKER_HEIGHT*0.5 for _, y in right_glyph])
        for _ in range(2):
            srs_source["color"].append("black")
            srs_source["name"].append("srs")
            srs_source['first_repeat'].append(left_pos)
            srs_source['second_repeat'].append(right_pos)
            srs_source["mutation_rate"].append(row["mutation_rate"])
            srs_source["line_width"].append(2)
            srs_source["line_color"].append("black")
            srs_source["sequence"].append(row["repeat"])
            srs_source["distance"].append(row["distance"])

    srs_glyphs = fig.patches(
        "x",
        "y",
        color="color",
        source=srs_source,
        alpha=0.5,
        line_color="line_color",
        line_width="line_width",
    )


    # Logic for highlighting
    outline_shape_left = [[0, 0], [1, 0], [1, MARKER_HEIGHT/2], [1, MARKER_HEIGHT], [0, MARKER_HEIGHT], [0, 0]]
    outline_shape_right = [[0, 0], [1, 0], [1, MARKER_HEIGHT], [0, MARKER_HEIGHT], [0, MARKER_HEIGHT/2], [0, 0]]
    line_shape = [[1, MARKER_HEIGHT/2], [1,MARKER_HEIGHT/2]]
    srs_outline_source = {
        "x": [],
        "y": [],
        "color": [],
        "line_color": [],
        "name": [],
        "first_repeat": [],
        "second_repeat": [],
        "mutation_rate": [],
        "line_width": [],
        "sequence": [],
        "distance": [],
    }
    srs_line_source = {
        "x": [],
        "y": [],
        "color": [],
    }
    for row in ssr_df.rows(named=True):
        left_pos = row["first_repeat"]
        right_pos = row["second_repeat"]
        repeat_len = row["repeat_len"]
        y_height = row["level"]*MARKER_HEIGHT*1.5
        if row["distance"] > OUTLINE_PADDING_X*4: # Distant SRSs
            left_outline = [[point[0]+left_pos, point[1]+OUTLINE_PADDING_Y] for point in outline_shape_left]
            left_outline_mods = [[-OUTLINE_PADDING_X, -OUTLINE_PADDING_Y],
                                 [OUTLINE_PADDING_X+repeat_len, -OUTLINE_PADDING_Y],
                                 [OUTLINE_PADDING_X*2+repeat_len, 0],
                                 [OUTLINE_PADDING_X+repeat_len, 0],
                                 [-OUTLINE_PADDING_X, 0],
                                 [-OUTLINE_PADDING_X, -OUTLINE_PADDING_Y]]
            for i in range(len(left_outline)):
                left_outline[i][0] = left_outline[i][0] + left_outline_mods[i][0]
                left_outline[i][1] = left_outline[i][1] + left_outline_mods[i][1] + y_height + MARKER_HEIGHT*0.5

            right_outline = [[point[0]+right_pos, point[1]+OUTLINE_PADDING_Y] for point in outline_shape_right]
            right_outline_mods = [[-OUTLINE_PADDING_X, -OUTLINE_PADDING_Y],
                                 [OUTLINE_PADDING_X+repeat_len, -OUTLINE_PADDING_Y],
                                 [OUTLINE_PADDING_X+repeat_len, 0],
                                 [-OUTLINE_PADDING_X+repeat_len, 0],
                                 [-OUTLINE_PADDING_X*2, 0],
                                 [-OUTLINE_PADDING_X, -OUTLINE_PADDING_Y]]
            for i in range(len(right_outline)):
                right_outline[i][0] = right_outline[i][0] + right_outline_mods[i][0]
                right_outline[i][1] = right_outline[i][1] + right_outline_mods[i][1] + y_height + MARKER_HEIGHT*0.5

            srs_outline_source["x"].append([x for x, _ in left_outline])
            srs_outline_source["y"].append([y for _, y in left_outline])
            srs_outline_source["x"].append([x for x, _ in right_outline])
            srs_outline_source["y"].append([y for _, y in right_outline])
            for _ in range(2):
                srs_outline_source["color"].append(None)
                srs_outline_source["name"].append("srs")
                srs_outline_source['first_repeat'].append(left_pos)
                srs_outline_source['second_repeat'].append(right_pos)
                srs_outline_source["mutation_rate"].append(row["mutation_rate"])
                srs_outline_source["line_width"].append(3)
                srs_outline_source["line_color"].append(color)
                srs_outline_source["sequence"].append(row["repeat"])
                srs_outline_source["distance"].append(row["distance"])


            line = deepcopy(line_shape)
            line[0][0] = left_outline[2][0]
            line[1][0] = right_outline[4][0]
            srs_line_source["x"].append([x for x, _ in line])
            srs_line_source["y"].append([y+OUTLINE_PADDING_Y+ y_height + MARKER_HEIGHT*0.5 for _, y in line])
            srs_line_source["color"].append(color)

        else: # Near SRSs
            l = [[point[0]+left_pos, point[1]+OUTLINE_PADDING_Y] for point in rmd_shape]
            r = [[point[0]+right_pos, point[1]+OUTLINE_PADDING_Y] for point in rmd_shape]

            outline = [l[0], r[1], r[2], l[3], l[4]]
            outline_mods = [[-OUTLINE_PADDING_X, -OUTLINE_PADDING_Y],
                            [OUTLINE_PADDING_X+repeat_len, -OUTLINE_PADDING_Y],
                            [OUTLINE_PADDING_X+repeat_len, OUTLINE_PADDING_Y],
                            [-OUTLINE_PADDING_X, OUTLINE_PADDING_Y],
                            [-OUTLINE_PADDING_X, -OUTLINE_PADDING_Y]]
            for i in range(len(outline)):
                outline[i][0] = outline[i][0] + outline_mods[i][0]
                outline[i][1] = outline[i][1] + outline_mods[i][1] + y_height + MARKER_HEIGHT*0.5
            srs_outline_source["x"].append([x for x, _ in outline])
            srs_outline_source["y"].append([y for _, y in outline])
            srs_outline_source["color"].append(None)
            srs_outline_source["name"].append("SRS")
            srs_outline_source['first_repeat'].append(left_pos)
            srs_outline_source['second_repeat'].append(right_pos)
            srs_outline_source["mutation_rate"].append(row["mutation_rate"])
            srs_outline_source["line_width"].append(3)
            srs_outline_source["line_color"].append(color)
            srs_outline_source["sequence"].append(row["repeat"])
            srs_outline_source["distance"].append(row["distance"])

    srs_glyphs_outline = fig.patches(
        "x",
        "y",
        color="color",
        source=srs_outline_source,
        alpha=0.5,
        line_color="line_color",
        line_width="line_width",
    )

    srs_lines = fig.multi_line(
        xs="x",
        ys="y",
        source=srs_line_source,
        alpha=0.5,
        color="color",
        line_width=3,
    )

    srs_glyphs_hover = HoverTool(
        renderers=[srs_glyphs_outline],
        tooltips=[
            ("Type", "@name"),
            ("Sequence", "@sequence"),
            ("First Repeat (left)", "@first_repeat"),
            ("Second Repeat (right)", "@second_repeat"),
            ("Distance", "@distance"),
            ("Mutation Rate", "@mutation_rate"),
        ],
    )

    fig.add_tools(srs_glyphs_hover)

    return fig

def plot_rmd(fig, rmd_df):
    height = 1
    rmd_shape = [[0, 0], [1, 0], [1, INNER_HEIGHT], [0, INNER_HEIGHT], [0, 0]]
    color = "red"

    # Logic for exact position
    rmd_source = {
        "x": [],
        "y": [],
        "color": [],
        "line_color": [],
        "name": [],
        "first_repeat": [],
        "second_repeat": [],
        "mutation_rate": [],
        "line_width": [],
        "sequence": [],
        "distance": [],
    }

    for row in rmd_df.rows(named=True):
        left_pos = row["first_repeat"]
        right_pos = row["second_repeat"]
        repeat_len = row["repeat_len"]
        y_height = row["level"]*MARKER_HEIGHT*1.5
        left_glyph = [[point[0]*repeat_len+left_pos, point[1]+OUTLINE_PADDING_Y] for point in rmd_shape]
        right_glyph = [[point[0]*repeat_len+right_pos, point[1]+OUTLINE_PADDING_Y] for point in rmd_shape]

        rmd_source["x"].append([x for x, _ in left_glyph])
        rmd_source["y"].append([y+y_height+MARKER_HEIGHT*0.5 for _, y in left_glyph])
        rmd_source["x"].append([x for x, _ in right_glyph])
        rmd_source["y"].append([y+y_height+ MARKER_HEIGHT*0.5 for _, y in right_glyph])
        for _ in range(2):
            rmd_source["color"].append("black")
            rmd_source["name"].append("rmd")
            rmd_source['first_repeat'].append(left_pos)
            rmd_source['second_repeat'].append(right_pos)
            rmd_source["mutation_rate"].append(row["mutation_rate"])
            rmd_source["line_width"].append(2)
            rmd_source["line_color"].append("black")
            rmd_source["sequence"].append(row["repeat"])
            rmd_source["distance"].append(row["distance"])

    rmd_glyphs = fig.patches(
        "x",
        "y",
        color="color",
        source=rmd_source,
        alpha=0.5,
        line_color="line_color",
        line_width="line_width",
    )


    # Logic for highlighting
    outline_shape_left = [[0, 0], [1, 0], [1, MARKER_HEIGHT/2], [1, MARKER_HEIGHT], [0, MARKER_HEIGHT], [0, 0]]
    outline_shape_right = [[0, 0], [1, 0], [1, MARKER_HEIGHT], [0, MARKER_HEIGHT], [0, MARKER_HEIGHT/2], [0, 0]]
    line_shape = [[1, MARKER_HEIGHT/2], [1,MARKER_HEIGHT/2]]
    rmd_outline_source = {
        "x": [],
        "y": [],
        "color": [],
        "line_color": [],
        "name": [],
        "first_repeat": [],
        "second_repeat": [],
        "mutation_rate": [],
        "line_width": [],
        "sequence": [],
        "distance": [],
    }
    rmd_line_source = {
        "x": [],
        "y": [],
        "color": [],
    }
    for row in rmd_df.rows(named=True):
        left_pos = row["first_repeat"]
        right_pos = row["second_repeat"]
        repeat_len = row["repeat_len"]
        y_height = row["level"]*MARKER_HEIGHT*1.5
        if row["distance"] > OUTLINE_PADDING_X*4: # Distant rmds
            left_outline = [[point[0]+left_pos, point[1]+OUTLINE_PADDING_Y] for point in outline_shape_left]
            left_outline_mods = [[-OUTLINE_PADDING_X, -OUTLINE_PADDING_Y],
                                 [OUTLINE_PADDING_X+repeat_len, -OUTLINE_PADDING_Y],
                                 [OUTLINE_PADDING_X*2+repeat_len, 0],
                                 [OUTLINE_PADDING_X+repeat_len, 0],
                                 [-OUTLINE_PADDING_X, 0],
                                 [-OUTLINE_PADDING_X, -OUTLINE_PADDING_Y]]
            for i in range(len(left_outline)):
                left_outline[i][0] = left_outline[i][0] + left_outline_mods[i][0]
                left_outline[i][1] = left_outline[i][1] + left_outline_mods[i][1] + y_height + MARKER_HEIGHT*0.5

            right_outline = [[point[0]+right_pos, point[1]+OUTLINE_PADDING_Y] for point in outline_shape_right]
            right_outline_mods = [[-OUTLINE_PADDING_X, -OUTLINE_PADDING_Y],
                                 [OUTLINE_PADDING_X+repeat_len, -OUTLINE_PADDING_Y],
                                 [OUTLINE_PADDING_X+repeat_len, 0],
                                 [-OUTLINE_PADDING_X, 0],
                                 [-OUTLINE_PADDING_X*2, 0],
                                 [-OUTLINE_PADDING_X, -OUTLINE_PADDING_Y]]
            for i in range(len(right_outline)):
                right_outline[i][0] = right_outline[i][0] + right_outline_mods[i][0]
                right_outline[i][1] = right_outline[i][1] + right_outline_mods[i][1] + y_height + MARKER_HEIGHT*0.5

            rmd_outline_source["x"].append([x for x, _ in left_outline])
            rmd_outline_source["y"].append([y for _, y in left_outline])
            rmd_outline_source["x"].append([x for x, _ in right_outline])
            rmd_outline_source["y"].append([y for _, y in right_outline])
            for _ in range(2):
                rmd_outline_source["color"].append(None)
                rmd_outline_source["name"].append("rmd")
                rmd_outline_source['first_repeat'].append(left_pos)
                rmd_outline_source['second_repeat'].append(right_pos)
                rmd_outline_source["mutation_rate"].append(row["mutation_rate"])
                rmd_outline_source["line_width"].append(3)
                rmd_outline_source["line_color"].append(color)
                rmd_outline_source["sequence"].append(row["repeat"])
                rmd_outline_source["distance"].append(row["distance"])


            line = deepcopy(line_shape)
            line[0][0] = left_outline[2][0]
            line[1][0] = right_outline[4][0]
            rmd_line_source["x"].append([x for x, _ in line])
            rmd_line_source["y"].append([y+OUTLINE_PADDING_Y+ y_height + MARKER_HEIGHT*0.5 for _, y in line])
            rmd_line_source["color"].append(color)

        else: # Near rmds
            l = [[point[0]+left_pos, point[1]+OUTLINE_PADDING_Y] for point in rmd_shape]
            r = [[point[0]+right_pos, point[1]+OUTLINE_PADDING_Y] for point in rmd_shape]

            outline = [l[0], r[1], r[2], l[3], l[4]]
            outline_mods = [[-OUTLINE_PADDING_X, -OUTLINE_PADDING_Y],
                            [OUTLINE_PADDING_X+repeat_len, -OUTLINE_PADDING_Y],
                            [OUTLINE_PADDING_X+repeat_len, OUTLINE_PADDING_Y],
                            [-OUTLINE_PADDING_X, OUTLINE_PADDING_Y],
                            [-OUTLINE_PADDING_X, -OUTLINE_PADDING_Y]]
            for i in range(len(outline)):
                outline[i][0] = outline[i][0] + outline_mods[i][0]
                outline[i][1] = outline[i][1] + outline_mods[i][1] + y_height + MARKER_HEIGHT*0.5
            rmd_outline_source["x"].append([x for x, _ in outline])
            rmd_outline_source["y"].append([y for _, y in outline])
            rmd_outline_source["color"].append(None)
            rmd_outline_source["name"].append("rmd")
            rmd_outline_source['first_repeat'].append(left_pos)
            rmd_outline_source['second_repeat'].append(right_pos)
            rmd_outline_source["mutation_rate"].append(row["mutation_rate"])
            rmd_outline_source["line_width"].append(3)
            rmd_outline_source["line_color"].append(color)
            rmd_outline_source["sequence"].append(row["repeat"])
            rmd_outline_source["distance"].append(row["distance"])

    rmd_glyphs_outline = fig.patches(
        "x",
        "y",
        color="color",
        source=rmd_outline_source,
        alpha=0.5,
        line_color="line_color",
        line_width="line_width",
    )

    rmd_lines = fig.multi_line(
        xs="x",
        ys="y",
        source=rmd_line_source,
        alpha=0.5,
        color="color",
        line_width=3,
    )

    rmd_glyphs_hover = HoverTool(
        renderers=[rmd_glyphs_outline],
        tooltips=[
            ("Type", "@name"),
            ("Sequence", "@sequence"),
            ("First Repeat (left)", "@first_repeat"),
            ("Second Repeat (right)", "@second_repeat"),
            ("Distance", "@distance"),
            ("Mutation Rate", "@mutation_rate"),
        ],
    )

    fig.add_tools(rmd_glyphs_hover)

    return fig



import polars as pl
import itertools
from ..constants import MIN_SRS_LEN, MAX_SHORT_SEQ_LEN


def _pairwise_slips(polars_df, column, is_circular, copies_cap=40) -> pl.DataFrame:
    """Recieve a polars dataframe with column of [List[type]]
    Returns the dataframe back with a pairwise list of positions"

    input:
        polars_df (dataframe): polars dataframe containing all repeats with lengths less than 16 base pairs
        column (str): name of column of [List[type]] in dataframe that contains the positions of each repeat
        is_circular (boolean): whether the scanned sequence is circular or not

    returns:
        polars dataframe containing all repeats with length less than 16 base pairs with a pairwise list of positions
    """

    # Shamelessly adapted from https://stackoverflow.com/questions/77354114/how-would-i-generate-combinations-of-items-within-polars-using-the-native-expres


    def map_function(list_o_things):
        return [
            sorted((thing_1, thing_2))
            for thing_1, thing_2 in itertools.combinations(list_o_things, 2)
        ]

    # Use linear pairing to find SSRs below the minimum SRS length (much faster)
    polars_df = polars_df.with_columns(
        pl.col('repeat').str.len_chars().alias('length')
    )

    linear_subset = polars_df.filter(pl.col('length') < MIN_SRS_LEN)
    pairwise = polars_df.filter(pl.col('length') >= MIN_SRS_LEN)

    high_mut_df = (
        pairwise.filter(
            pl.col("length") == MAX_SHORT_SEQ_LEN-1,
            pl.col("position").list.len() >= copies_cap
            )
    )

    # terminate if there are 40+ occurrences of a single repeat
    if len(high_mut_df) > 0:
        raise ValueError("This sequence is highly mutagenic. Stopping execution")

    linear_subset = _linear_slips(linear_subset, column, is_circular=is_circular)

    # Run pairwise for SRS's above minimum length
    pairwise = pairwise.lazy().with_columns(
        pl.col(column)
        .map_elements(map_function, return_dtype=pl.List(pl.List(pl.Int64)))
        .alias(f"pairings").cast(pl.List(pl.List(pl.Int32)))
    )

    linear_subset = linear_subset
    pairwise = pairwise.collect().select(pl.col('repeat'), pl.col('pairings'))

    pairwise = pl.concat([linear_subset, pairwise])

    return pairwise


def _linear_slips(polars_df, column, is_circular=False) -> pl.LazyFrame:
    """Recieve a polars dataframe with column of [List[type]]
    Returns the dataframe back with a linear list of pairings

    input:
        polars_df (dataframe): polars dataframe containing all repeats with lengths less than 16 base pairs
        column (str): name of column of [List[type]] in dataframe that contains the positions of each repeat
        is_circular (boolean): whether the scanned sequence is circular or not

    returns:
        polars dataframe containing all repeats with length less than 16 base pairs with a linear list of pairings
    """

    nrows = polars_df.select(pl.len()).item()

    linear_df = (
        polars_df.with_columns(instances=pl.col(column).list.len())
        .with_columns(
            duplicate_column=pl.col(column)
            .list.tail(pl.col(column).list.len() - 1)
            .list.concat(pl.col(column).list.first())
        )
        .explode([column, "duplicate_column"])
        .select(
            [
                "repeat",
                pl.struct([column, "duplicate_column"]).alias("pairings"),
            ]
        )
        .group_by("repeat")
        .agg(pl.col("pairings"))
    )

    if not is_circular:
        linear_df = linear_df.with_columns(
            pairings=pl.col("pairings").list.head(pl.col("pairings").list.len() - 1)
        )

    linear_df = (
        linear_df.explode("pairings")
        .unnest("pairings")
        .select(
            pl.col("repeat"),
            pl.concat_list(pl.col("position"), pl.col("duplicate_column"))
            .list.sort()
            .alias("pairings"),
        )
        .group_by("repeat")
        .agg(pl.col("pairings").unique())
    )

    return linear_df



import polars as pl

def _collapse_ssr(polars_df) -> pl.DataFrame:
    """Takes in dataframe of SSRs, returns dataframe of SSRs collapsed down.

    input:
        polars_df: polars datafrmae containing all repeats found, with a column for repeat type

    returns:
        polars dataframe containing only the SSRs found
    """

    collapsed_ssrs = (
        polars_df.filter(pl.col("category") == "SSR")
        .select(["repeat", "repeat_len", "pairings", "wraparound"])
        .lazy()
        # Collect the positions from all potentially participating SSRs
        .explode("pairings")
        .group_by(["repeat", "repeat_len"])
        .agg("pairings", "wraparound")
        .with_columns(
            positions=pl.col("pairings").list.unique().list.sort(),
        )
        .with_columns(differences=pl.col("positions").list.diff())
        .collect()
    )
    # Somehow couldnt figure out how to do this in pure polars

    # Identify start positions. If distance!=0, it a start of a repeat
    collapsed_ssrs = collapsed_ssrs.to_pandas()
    collapsed_ssrs["differences"] = (
        collapsed_ssrs["differences"] - collapsed_ssrs["repeat_len"]
    )
    collapsed_ssrs = pl.from_pandas(collapsed_ssrs)

    collapsed_ssrs = (
        collapsed_ssrs
        .with_columns(
            pl.col("differences").cast(pl.List(pl.Float64))
            )
        .with_columns(
            truth_table=(
                pl.col("differences").list.eval(
                    (pl.element() != 0).or_(pl.element().is_null())
                )
            )
        )
    )

    # Apply the truth table
    collapsed_ssrs = collapsed_ssrs.to_pandas()
    collapsed_ssrs["starts"] = (collapsed_ssrs["truth_table"]) * (
        collapsed_ssrs["positions"] + 1
    )

    # Fill out
    collapsed_ssrs = pl.from_pandas(collapsed_ssrs)
    collapsed_ssrs = collapsed_ssrs.with_columns(
        starts=pl.col("starts")
        .explode()
        .replace(0, None)
        .forward_fill()
        .implode()
        .over("repeat")
    )

    # We had to apply an offeset with the truth table
    # To prevent bp-0 start positions from nulling out. This undoes that.
    collapsed_ssrs = collapsed_ssrs.to_pandas()
    collapsed_ssrs["starts"] = collapsed_ssrs["starts"] - 1
    collapsed_ssrs = pl.from_pandas(collapsed_ssrs).select(
        pl.col(["repeat", "repeat_len", "starts", "wraparound"])
    )

    # Count repeats from each start position
    collapsed_ssrs = (
        (
            collapsed_ssrs.lazy()
            .explode("starts")
            .group_by("repeat", "repeat_len", "starts", "wraparound")
            .count()
        )
        .rename({"starts": "start"})
        .collect()
    )

    # Correct for circular
    if collapsed_ssrs.height > 0:
        # Correct for circular SSR
        collapsed_ssrs = (
            collapsed_ssrs
            .sort("start")
            .group_by("repeat", "repeat_len")
            .agg(
                pl.col("start"),
                pl.col("count"),
                pl.first("wraparound")
            )
            .with_columns(
                pl.when(pl.col("wraparound").list.contains(True))
                .then(
                    pl.concat_list([
                        pl.col("count").list.slice(1, pl.col("count").list.len() - 2),
                        (pl.col("count").list.first() + pl.col("count").list.last())
                    ]).alias("count")
                )
                .otherwise(pl.col("count")),
                pl.when(pl.col("wraparound").list.contains(True))
                .then(pl.col("start").list.slice(1).alias("start"))
                .otherwise(pl.col("start"))
            )
            .explode(
                pl.col("start"),
                pl.col("count")
            )
        )

    return collapsed_ssrs



from ..constants import MIN_SHORT_SEQ_LEN, MAX_SHORT_SEQ_LEN
import polars as pl

def collect_subsequences(seq, isCircular, window_max=16) -> pl.LazyFrame:
    """Scans across a given input sequence and returns a list of subsequences

    input:
        seq (string): the sequence to be scanned
        isCircular (boolean): Whether the sequence is circular or not

    returns:
        polars dataframe containing every repeat smaller than 16 base pairs in the input sequence

    """

    seq_len = len(seq)

    if isCircular:
        # adds first 20 bp to end
        seq = seq + seq[0:20]

    def scan_genome():
        # Probably room for optimizations here
        for i, _ in enumerate(seq[:seq_len]):
            for j in range(MIN_SHORT_SEQ_LEN, MAX_SHORT_SEQ_LEN):
                if len(seq[i : i + j]) > MIN_SHORT_SEQ_LEN:
                    sub_seq = seq[i : i + j]
                    yield {"repeat": str(sub_seq), "position": i}

    repeats = (
        pl.LazyFrame(scan_genome()).group_by("repeat").agg(pl.col("position")).collect()
    ).cast({"position": pl.List(pl.Int32)})

    return repeats


def highly_mut(df: pl.DataFrame):
    df = (
        df.filter(pl.col("repeat_len") == MAX_SHORT_SEQ_LEN-1)
        .group_by(pl.col("repeat"))
        .agg(
            pl.col("pairings"), 
            pl.first("repeat_len")
            )
        .with_columns(pl.col("pairings").list.unique().alias("positions"))
        .filter(pl.col("positions").list.len() >= 40)
    )

    # terminate if there are 40+ occurrences of a single repeat
    if df.is_empty():
        return False
    else:
        return True




def _scan_RMD(df: pl.DataFrame, seq, seq_len, isCircular) -> pl.DataFrame:
    """Scans for RMDs

    input:
        df (dataframe): dataframe containing all repeats smaller than 16 base pairs
        seq (string): the sequence to be scanned
        seq_len (int): the length of the sequence
        isCircular (boolean): whether the sequence is circular or not

    returns:
        polars dataframe containing every repeat in the input sequence

    """

    known_long_repeats = df.filter(pl.col("repeat_len") == (MAX_SHORT_SEQ_LEN - 1))
    RMD_df = pl.DataFrame(
        {
            "repeat": pl.Series("repeat", [], pl.Utf8),
            "pairings": pl.Series("pairings", [], pl.List(pl.Int64)),
            "repeat_len": pl.Series("repeat_len", [], pl.Int64),
        }
    )

    def check_larger_repeats(positions, seq):
        completed = False
        step = 50
        length = MAX_SHORT_SEQ_LEN
        pos1 = positions[0]
        pos2 = positions[1]
        largest = False
        wrap = 0

        while not completed:
            prvlength = length
            length += step
            # already know they are 15 bp repeats
            if prvlength < 16:
                prvlength = 16

            # pos2 is always after pos1
            if pos2 + length > seq_len:
                if isCircular:
                    wrap += step
                    seq = seq + seq[wrap - step : wrap]
                else:
                    length = seq_len - pos2
                    largest = True
            # largest possible repeat is seq_len/2 bp
            if length >= int(seq_len / 2):
                largest = True
                length = int(seq_len / 2)
            # prevent out of bound error
            if len(seq) - pos2 < length:
                length = len(seq) - pos2

            # if sequences not equal, then largest repeat has been passed
            if (seq[pos1 : pos1 + length] != seq[pos2 : pos2 + length]) or (largest):
                # iterate 1 by 1.
                # Uses length+1 because in range stops before last index
                for j in range(prvlength, length + 1):
                    # uses j-1 because substrings end before last index
                    if seq[pos1 + (j - 1)] == seq[pos2 + (j - 1)]:
                        sub_seq = seq[pos1 : pos1 + j]
                        yield {
                            "repeat": str(sub_seq),
                            "pairings": [pos1, pos2],
                            "repeat_len": j,
                        }
                    else:
                        break
                completed = True

    def store_RMD(positions, seq):
        nonlocal RMD_df
        repeats = pl.DataFrame(check_larger_repeats(positions, seq))

        # if larger repeats were found, then repeats will have 3 columns ("repeat", "pairings", "repeat_len")
        if repeats.width == 3:
            RMD_df = pl.concat([RMD_df, repeats])
        return None

    # Apply the function to the DataFrame
    known_long_repeats.with_columns(
        pl.col("pairings").map_elements(
            lambda pairings: store_RMD(pairings, seq), return_dtype=pl.List(pl.Null)
        )
    )
    RMD_df = RMD_df.with_columns(
        pl.col("pairings").cast(pl.List(pl.Int32)),
        pl.col("repeat_len").cast(pl.Int32))

    df = pl.concat([df, RMD_df])

    return df  # In the same format as df alongside the original data



import pathlib
import logging
import csv
import hashlib
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from pathlib import Path
from ..constants import FASTA_EXTS, GBK_EXTS
from .EFMSequence import EFMSequence
from .bad_state_mitigation import BadSequenceError, detect_special_cases

logger = logging.getLogger(__name__)
logging.getLogger(__name__).addHandler(logging.NullHandler())

def parse_file(filepath: pathlib.Path, use_filename: bool = True, iscircular = False) -> list:
    """
    parses the inputted files and returns a list of sequences found in each file

    input:
        filepath: path to the multifasta, genbank, or csv file containing the sequences to be scanned

    returns:
        list containing all the sequences to be scanned
    """

    path_as_string = str(filepath)
    if not filepath.exists():
        raise OSError("File {} does not exist.".format(path_as_string))
    elif filepath.suffix in FASTA_EXTS:
        sequences = SeqIO.parse(path_as_string, "fasta")
    elif filepath.suffix in GBK_EXTS:
        sequences = SeqIO.parse(path_as_string, "genbank")
    elif filepath.suffix in [".csv"]:
        sequences = parse_csv(path_as_string)
    else:
        raise ValueError(
            f"File {filepath} is not a known file format. Must be one of {FASTA_EXTS + GBK_EXTS + [".csv"]}."
        )

    # If the genbank file doesnt have a name, add it
    test_sequences = []
    with open(filepath) as f:
        fileinfo = f.read().encode()
        filehash = hashlib.md5(fileinfo)
    for i, seq in enumerate(sequences):
        if use_filename:
            filename = Path(filepath).stem
            if not seq.name:
                seq.name = f"{filename}"
            if not seq.description or seq.description == '':
                seq.description = f"{filename}"
        originhash = filehash.hexdigest() + hashlib.md5(bytes(i) + str(iscircular).encode()).hexdigest()
        efmseq = EFMSequence(seq, iscircular, originhash)
        test_sequences.append(efmseq)
    return test_sequences


def validate_sequences(sequences, circular=True, max_len=None):
    cumulative_length = 0
    for seq in sequences:
        cumulative_length += len(seq)
        if max_len and cumulative_length > max_len:
            raise BadSequenceError(
                f"Input sequence(s) is too long. Max length is {max_len} bases."
            )
        validate_sequence(seq)

def validate_sequence(seq):
    IUPAC_BASES = set("ACGTURYSWKMBDHVNacgturyswkmbdhvn")
    sequence = str(seq.seq)
    seq_set = set(sequence.upper().replace(" ", ""))
    if sequence is "":
        raise BadSequenceError(
            f"Input contains an empty sequence."
        )
    if not seq_set.issubset(IUPAC_BASES):
        raise BadSequenceError(
            f"Input sequence contains invalid characters. Only IUPAC bases are allowed."
        )
    if "t" in seq_set and "u" in seq_set:
        raise BadSequenceError(
            f"Input sequence has both T and U. This is probably a mistake."
        )
    if seq_set.issubset("RYSWKMBDHVN"):
        raise BadSequenceError(
            f"EFM Calculator cannot currently handle IUPAC ambiguity codes."
        )
    detect_special_cases(seq)

def parse_csv(path_as_string):
    csv.field_size_limit(100000000)
    with open(path_as_string, "r") as csvfile:
        csvreader = csv.reader(csvfile)
        headers = csvreader.__next__()
        if not headers:
            raise BadSequenceError("CSV file is empty or malformed")
        try:
            names = headers.index("name")
        except:
            names = None
        try:
            seqs = headers.index("seq")
        except:
            raise BadSequenceError(
                f"CSV file has no 'seq' column"
            )
        sequences = []

        try:
            for i, entry in enumerate(csvreader):
                if names is not None:
                    name = entry[names]
                else:
                    name = f"sequence"
                sequence = entry[seqs]
                formatted_entry = SeqRecord(Seq(sequence),name=name,description=name)
                sequences.append(formatted_entry)
                csvreader = csv.reader(csvfile)
        except csv.Error as e: # Should be handled better probably
            raise e

    return sequences



import logging
import polars as pl
from collections import namedtuple
from collections import Counter, defaultdict
from rich import print
import Bio
import tempfile
from typing import List
from ..constants import MIN_SHORT_SEQ_LEN, MAX_SHORT_SEQ_LEN, UNKNOWN_REC_TYPE, SUB_RATE, MIN_SRS_LEN
from ..utilities import FakeBar

from .subsequence_curation import collect_subsequences, _scan_RMD, highly_mut
from .detection_strats import _pairwise_slips, _linear_slips
from .classify_ssr import _collapse_ssr
from .classify_srs_rmd import _calculate_distances, _categorize_efm


logger = logging.getLogger(__name__)
logging.getLogger(__name__).addHandler(logging.NullHandler())


def predict(seq: str, strategy: str, isCircular: bool) -> List[pl.DataFrame]:
    """Scans and predicts SSRs and RMDs. Returns dataframes representing each

    input:
        seq (string): the sequence to be scanned
        strategy (str): the scanning strategy to be used. Either pairwise or linear
        isCircular (boolean): whether the sequence is circular or not

    returns:
        ssr_df (dataframe): dataframe containing all the SSRs found in the sequence
        srs_df (dataframe): dataframe containing all the SRSs found in the sequence
        rmd_df (dataframe): dataframe containing all the RMDs found in the sequence

    """
    seq = seq.upper().replace(" ", "")
    seq_len = len(seq)

    valid_strategies = ["pairwise", "linear"]
    if strategy not in valid_strategies:
        raise ValueError(
            f"Invalid strategy: {strategy}. Must be one of {valid_strategies}"
        )

    # Curate target sequences

    repeat_df = collect_subsequences(seq, isCircular)

    # Filter out sequences
    repeat_df = repeat_df.filter(pl.col("position").list.len() > 1)
    num_repeated_sequences = repeat_df.select(pl.len()).item()
    # -- end debugging

    # Create list of positions
    if strategy == "pairwise":
        repeat_df = _pairwise_slips(repeat_df, "position", isCircular)
    elif strategy == "linear":
        repeat_df = _linear_slips(repeat_df, "position", isCircular)
    else:
        raise ValueError("Invalid strategy")

    if "position" in repeat_df:
        repeat_df = repeat_df.drop("position")
    if "position_corrected" in repeat_df:
        repeat_df = repeat_df.drop("position_corrected")
    repeat_df = repeat_df.explode("pairings")
    # Get length of each repeat
    repeat_df = repeat_df.with_columns(
        pl.col("repeat").str.len_chars().alias("repeat_len").cast(pl.Int32)
    )





    # Upgrade long SRSs to RMDs
    repeat_df = _scan_RMD(repeat_df, seq, seq_len, isCircular)

    # Calculate Distances
    repeat_df = _calculate_distances(repeat_df, seq_len, isCircular)
    repeat_df = repeat_df.filter(pl.col("distance") >= 0)
    repeat_df = repeat_df.unique()

    # Categorize positions
    repeat_df = _categorize_efm(repeat_df)
    # Collapse SSRs down
    ssr_df = _collapse_ssr(repeat_df).select(
        pl.col(["repeat", "repeat_len", "start", "count"])
    )

    # Process and Split SRS and RMD

    repeat_df = repeat_df.lazy().filter(pl.col("category") != "SSR").collect()

    # Remove SRS that are shorter than min_ssr_length
    repeat_df.filter(pl.col("repeat_len") >= MIN_SRS_LEN)

    if len(repeat_df) > 0:
        repeat_df = (
            repeat_df.lazy()
            .select(
                pl.col(["repeat", "repeat_len", "pairings", "distance", "category"])
            )
            .collect()
            .rechunk()  # Weird issue with invalid pairing state
            .lazy()
            .with_columns(
                pl.col("pairings").list.to_struct(
                    fields=[
                        "first_repeat",
                        "second_repeat",
                    ]
                )
            )
            .unnest("pairings")
        ).collect()
    else:
        schema = {
            "repeat": pl.Utf8,
            "repeat_len": pl.Int32,
            "first_repeat": pl.Int32,
            "second_repeat": pl.Int32,
            "distance": pl.Int32,
            "category": pl.Categorical,
        }
        repeat_df = pl.DataFrame(schema=schema)

    srs_df = repeat_df.filter(pl.col("category") == "SRS").select(
        pl.col(["repeat", "repeat_len", "first_repeat", "second_repeat", "distance"])
    )
    rmd_df = repeat_df.filter(pl.col("category") == "RMD").select(
        pl.col(["repeat", "repeat_len", "first_repeat", "second_repeat", "distance"])
    )

    return [ssr_df, srs_df, rmd_df]



import polars as pl


def eval_top(ssr_df=None, srs_df=None, rmd_df=None, num_report: int = 10):
    # Get the top n=num_report contributors across all categories
    valid_dataframes = []
    if isinstance(ssr_df, pl.DataFrame) and not ssr_df.is_empty():
        ssr_df = ssr_df.sort(by="mutation_rate")
        ssr_df_top = (
            ssr_df.with_columns(source=pl.lit("SSR"))
            .sort(by="mutation_rate", descending=True)
            .select(pl.col(["repeat", "source", "mutation_rate", "predid"]))
            .head(num_report)
        )
        valid_dataframes.append(ssr_df_top)

    if isinstance(srs_df, pl.DataFrame) and not srs_df.is_empty():
        srs_df = srs_df.sort(by="mutation_rate")
        srs_df_top = (
            srs_df.with_columns(source=pl.lit("SRS"))
            .sort(by="mutation_rate", descending=True)
            .select(pl.col(["repeat", "source", "mutation_rate", "predid"]))
            .head(num_report)
        )
        valid_dataframes.append(srs_df_top)

    if isinstance(srs_df, pl.DataFrame) and not ssr_df.is_empty():
        rmd_df = rmd_df.sort(by="mutation_rate")
        rmd_df_top = (
            rmd_df.with_columns(source=pl.lit("RMD"))
            .sort(by="mutation_rate", descending=True)
            .select(pl.col(["repeat", "source", "mutation_rate", "predid"]))
            .head(num_report)
        )
        valid_dataframes.append(rmd_df_top)

    if len(valid_dataframes) > 0:
        merged_df = pl.concat(valid_dataframes)
        merged_df = merged_df.sort(by="mutation_rate", descending=True).head(num_report)
    # If no SSR, SRS, or RMD are found
    else:
        merged_df = srs_df

    return merged_df



import logging
import polars as pl
import tempfile
import itertools
from progress.spinner import Spinner
from progress.bar import IncrementalBar
import numpy as np
from ..constants import MAX_SHORT_SEQ_LEN, MIN_SHORT_SEQ_LEN


logger = logging.getLogger(__name__)
logging.getLogger(__name__).addHandler(logging.NullHandler())


def _calculate_distances(polars_df, seq_len, circular) -> pl.LazyFrame:
    """
    calculates the distance between occurrences of each repeat

    input:
        polars_df: polars dataframe containing all repeats found in the scanned sequence
        seq_len (int): length of the scanned sequence
        circular: whether the scanned sequence is circular or not

    returns:
        inputted polars dataframe with a column of distances between each repeat
    """
    distance_df = polars_df.with_columns(
        distance=pl.col("pairings").list.diff().list.get(1) - pl.col("repeat_len").cast(pl.Int32)
    )
    if circular:
         distance_df = (
            distance_df
            .with_columns((pl.col("pairings").list.get(0) - (pl.col("pairings").list.get(1) + pl.col("repeat_len") - seq_len)).alias("alt_len"))
            .with_columns(
                distance=pl.when(pl.col("alt_len") >= 0)
                .then(
                    pl.when(pl.col("distance") > pl.col("alt_len"))
                    .then(pl.col("alt_len"))
                    .otherwise(pl.col("distance"))
                )
                .otherwise(pl.col("distance")).cast(pl.Int32),
                wraparound=pl.when(pl.col("alt_len") >= 0)
                .then(
                    pl.when(pl.col("distance") > pl.col("alt_len"))
                    .then(True)
                    .otherwise(False)
                )
                .otherwise(False)
                )
        )
    else:
        distance_df = distance_df.with_columns(
            wraparound = False,
        )
    return distance_df

def _categorize_efm(polars_df) -> pl.DataFrame:
    """
    categorizes every repeat as SSR, SRS, or RMD

    input:
        polars_df: polars dataframe containing all repeats found in the scanned sequence

    returns:
        inputted polars dataframe with a column for repeat type, either SSR, SRS, or RMD
    """
    categorized_df = polars_df.with_columns(
        category=pl.when(pl.col("distance") > 0)
        .then(
            pl.when(pl.col("repeat_len") > MIN_SHORT_SEQ_LEN)
            .then(
            pl.when(pl.col("repeat_len") < MAX_SHORT_SEQ_LEN)
            .then(pl.lit("SRS"))
            .otherwise(pl.lit("RMD"))
            )
        )
        .otherwise(pl.lit("SSR"))
        .cast(pl.Categorical)
    )
    return categorized_df



"""Manages sequence-specific state information for the webapp"""
import polars as pl
from .vis_utils import eval_top
import streamlit as st
import logging
for name, l in logging.root.manager.loggerDict.items():
    if "streamlit" in name:
        l.disabled = True

from ..ingest.EFMSequence import sequence_to_features_df
from st_aggrid import AgGrid, GridOptionsBuilder, GridUpdateMode, JsCode

class SequenceState():
    """Manages sequence-specific state information for the webapp"""
    def __init__(self, efmsequence):
        self.efmsequence = efmsequence


        self._ssr_webapp_state = None

        self._webapp_ssrs = None
        self._webapp_srss = None
        self._webapp_rmds = None
        self._webapp_top = None

        self._unique_annotations = {}

        self.ready = False
        self.refreshed = False
        self.counter = 0

        self.topkey = f"toptable{self.counter}"
        self.ssrkey = f"ssrtable{self.counter}"
        self.srskey = f"srstable{self.counter}"
        self.rmdkey = f"rmdtable{self.counter}"

    # Data from upstream EFMSequence

    @property
    def predicted(self):
        return self.efmsequence.predicted

    @property
    def unique_annotations(self):
        return self.efmsequence.unique_annotations

    @property
    def seq(self):
        return self.efmsequence.seq

    @property
    def features(self):
        return self.efmsequence.features

    def post_predict_processing(self):
        if not self.predicted:
            raise ValueError("Predictions have not yet been completed")
        elif self.ready:
            return
        self._webapp_ssrs = self.efmsequence._ssrs.sort(by="mutation_rate", descending=True)
        self._webapp_srss = self.efmsequence._srss.sort(by="mutation_rate", descending=True)
        self._webapp_rmds = self.efmsequence._rmds.sort(by="mutation_rate", descending=True)
        self._webapp_top = self.efmsequence._top.sort(by="mutation_rate", descending=True)
        self._shown_predictions = [x[0] for x in self.efmsequence._top.select(pl.col("predid")).unique().rows()]
        self._filtered_ssrs = None
        self._filtered_srss = None
        self._filtered_rmds = None
        self._filtered_top = None
        self._last_filters = []
        self.set_filters(None)

        self.last_top_selections = None
        self.last_ssr_selections = None
        self.last_srs_selections = None
        self.last_rmd_selections = None

    def reset_selected_predictions(self):
        self._shown_predictions = [x[0] for x in self._filtered_top.select(pl.col("predid")).unique().rows()]

    @property
    def unique_annotations(self):
        if not self.efmsequence.annotations:
            return None
        if isinstance(self._unique_annotations, dict) and self._unique_annotations == {}:
            self._unique_annotations = sequence_to_features_df(self.efmsequence, self.efmsequence.is_circular)
            self._unique_annotations = self._unique_annotations.with_columns(
                pl.concat_str([pl.col("annotations"), pl.lit(" ("), pl.col("left_bound"), pl.lit("-"), pl.col("right_bound"), pl.lit(")")]).alias("annotationobjexpanded_names")
            )
        unique_expaned_names = self._unique_annotations.select(pl.col("annotationobjexpanded_names")).unique().rows()
        unique_expaned_names = [x[0] for x in unique_expaned_names]

        return sorted(unique_expaned_names)

    # Data pertainint to webapp
    def set_filters(self, annotations):
        if annotations:
            annotation_objects = self._unique_annotations.filter(pl.col("annotationobjexpanded_names").is_in(annotations))
            annotation_objects = annotation_objects.select(pl.col("annotationobjects")).unique().rows()
            annotation_objects = [x[0] for x in annotation_objects]
            self._filtered_ssrs = self._webapp_ssrs.filter(pl.col("annotationobjects").list.set_intersection(annotation_objects).list.len() != 0)
            self._filtered_srss = self._webapp_srss.filter(pl.col("annotationobjects").list.set_intersection(annotation_objects).list.len() != 0)
            self._filtered_rmds = self._webapp_rmds.filter(pl.col("annotationobjects").list.set_intersection(annotation_objects).list.len() != 0)
            self._filtered_top = eval_top(self._filtered_ssrs, self._filtered_srss, self._filtered_rmds).with_columns(pl.lit(False).alias("show"))
        else:
            self._filtered_ssrs = self._webapp_ssrs
            self._filtered_srss = self._webapp_srss
            self._filtered_rmds = self._webapp_rmds
            self._filtered_top = self._webapp_top

    @property
    def top_webapp_table(self):
        """Streamlit aggrid table representing ssr data"""

        pandas_conversion = self._filtered_top.to_pandas()
        return AgGrid(pandas_conversion,
                            gridOptions=self._top_webapp_state,
                            height=500,
                            fit_columns_on_grid_load=True,
                            allow_unsafe_jscode = True,
                            update_mode = GridUpdateMode.SELECTION_CHANGED,
                            key = self.topkey,
                            callback = self.update_top_session)


    def rebuild_top_table(self):
        cell_hover_handler = JsCode("""""")
        js_hover_handler = """"""

        preselected_indices = self.get_shown_predictions(self._filtered_top)
        preselected_indices = [str(index) for index in preselected_indices]
        if self.last_top_selections is None:
            self.last_top_selections = preselected_indices

        pandas_conversion = self._filtered_top.to_pandas()
        builder = GridOptionsBuilder.from_dataframe(pandas_conversion)
        builder.configure_selection(selection_mode='multiple', use_checkbox= True, pre_selected_rows= preselected_indices)

        builder.configure_grid_options(onCellMouseOver=cell_hover_handler)
        builder.configure_column("repeat", header_name="Sequence", tooltipField="repeat")
        builder.configure_column("mutation_rate", header_name="Mutation Rate",
                    type=["numericColumn"], valueFormatter="x.toExponential(2)")
        builder.configure_column("annotations", header_name="Annotations", tooltipField="annotations")
        builder.configure_column("predid", hide = True)
        builder.configure_column("annotationobjects", hide = True)
        grid_options = builder.build()
        self._top_webapp_state = grid_options

    def update_top_session(self, callbackobj):
        self.refresh_last_shown()
        state = callbackobj.grid_response["gridState"]
        new_selection = state.get('rowSelection', [])

        if self.last_top_selections is not None and self.last_top_selections == new_selection:
            return

        if self.last_top_selections is None:
            self.last_top_selections = []
        if self.last_top_selections == new_selection:
            return

        self.counter += 1
        self.ssrkey = f"ssrtable{self.counter}"
        self.srskey = f"srstable{self.counter}"
        self.rmdkey = f"rnmdable{self.counter}"

        to_drop = [int(x) for x in self.last_top_selections if x not in new_selection]
        to_add = [int(x) for x in new_selection if x not in self.last_top_selections]


        self._update_general_table(self._filtered_top, to_drop, to_add)

        self.last_top_selections = new_selection


    # SSR Table Functions
    @property
    def ssr_webapp_table(self):
        """Streamlit aggrid table representing ssr data"""

        pandas_conversion = self._filtered_ssrs.to_pandas()
        return AgGrid(pandas_conversion,
                            gridOptions=self._ssr_webapp_state,
                            height=500,
                            fit_columns_on_grid_load=True,
                            allow_unsafe_jscode = True,
                            update_mode = GridUpdateMode.SELECTION_CHANGED,
                            key=self.ssrkey,
                            callback=self.update_ssr_session,
                            reload_data=False)

    def rebuild_ssr_table(self):
        cell_hover_handler = JsCode("""
                    function(params) {
                        // debug
                        const clickedColumn = params.column.colId;
                        const clickedRowIndex = params.rowIndex;
                        const clickedValue = params.node.data[clickedColumn];

                        const predidValue = params.node.data["predid"];

                        // Display information about the click
                        const message = `You hovered on row ${clickedRowIndex}, column ${clickedColumn}, value is ${predidValue}`;
                        console.log(message);

                        window.parent.postMessage(
                            {type: 'cellMouseOver', predid: predidValue}, '*'
                            );

                    }
                    """)
        js_hover_handler = """
                        window.addEventListener("message", (event) => {
                            if (event.data.type === "cellMouseOver") {
                                resolve(event.data.predid);
                            }
                        });
                    """

        preselected_indices = self.get_shown_predictions(self._filtered_ssrs)
        preselected_indices = [str(index) for index in preselected_indices]
        if self.last_ssr_selections is None:
            self.last_ssr_selections = preselected_indices

        pandas_conversion = self._filtered_ssrs.to_pandas()
        builder = GridOptionsBuilder.from_dataframe(pandas_conversion)
        builder.configure_selection(selection_mode='multiple', use_checkbox= True, pre_selected_rows= preselected_indices)

        builder.configure_grid_options(onCellMouseOver=cell_hover_handler)
        builder.configure_column("repeat", header_name="Sequence", tooltipField="repeat")
        builder.configure_column("repeat_len", header_name="Repeat Length", type=["numericColumn"])
        builder.configure_column("start", header_name="Start", type=["numericColumn"])
        builder.configure_column("count", header_name="Count", type=["numericColumn"])
        builder.configure_column("mutation_rate", header_name="Mutation Rate",
                    type=["numericColumn"], valueFormatter="x.toExponential(2)")
        builder.configure_column("annotations", header_name="Annotations", tooltipField="annotations")
        builder.configure_column("predid", hide = True)
        builder.configure_column("annotationobjects", hide = True)
        grid_options = builder.build()
        self._ssr_webapp_state = grid_options

    def update_ssr_session(self, callbackobj):
        self.refresh_last_shown()
        state = callbackobj.grid_response["gridState"]
        new_selection = state.get('rowSelection', [])

        if self.last_ssr_selections is not None and self.last_ssr_selections == new_selection:
            return
        if self.last_ssr_selections is None:
            self.last_ssr_selections = []
        if self.last_ssr_selections == new_selection:
            return

        self.counter += 1
        self.topkey = f"toptable{self.counter}"

        to_drop = [int(x) for x in self.last_ssr_selections if x not in new_selection]
        to_add = [int(x) for x in new_selection if x not in self.last_ssr_selections]


        self._update_general_table(self._filtered_ssrs, to_drop, to_add)

        self.last_ssr_selections = new_selection

    # SRS Table Functions
    @property
    def srs_webapp_table(self):
        """Streamlit aggrid table representing srs data"""

        pandas_conversion = self._filtered_srss.to_pandas()
        return AgGrid(pandas_conversion,
                            gridOptions=self._srs_webapp_state,
                            height=500,
                            fit_columns_on_grid_load=True,
                            allow_unsafe_jscode = True,
                            update_mode = GridUpdateMode.SELECTION_CHANGED,
                            key=self.srskey,
                            callback=self.update_srs_session,
                            reload_data=False)

    def rebuild_srs_table(self):
        cell_hover_handler = JsCode("""
                    function(params) {
                        // debug
                        const clickedColumn = params.column.colId;
                        const clickedRowIndex = params.rowIndex;
                        const clickedValue = params.node.data[clickedColumn];

                        const predidValue = params.node.data["predid"];

                        // Display information about the click
                        const message = `You hovered on row ${clickedRowIndex}, column ${clickedColumn}, value is ${predidValue}`;
                        console.log(message);

                        window.parent.postMessage(
                            {type: 'cellMouseOver', predid: predidValue}, '*'
                            );

                    }
                    """)
        js_hover_handler = """

                    """

        preselected_indices = self.get_shown_predictions(self._filtered_srss)
        preselected_indices = [str(index) for index in preselected_indices]
        if self.last_srs_selections is None:
            self.last_srs_selections = preselected_indices

        pandas_conversion = self._filtered_srss.to_pandas()
        builder = GridOptionsBuilder.from_dataframe(pandas_conversion)
        builder.configure_selection(selection_mode='multiple', use_checkbox= True, pre_selected_rows= preselected_indices)
        builder.configure_grid_options(onCellMouseOver=cell_hover_handler)
        builder.configure_column("repeat", header_name="Sequence", tooltipField="repeat")
        builder.configure_column("repeat_len", header_name="Repeat Length", type=["numericColumn"])
        builder.configure_column("first_repeat", header_name="First Repeat", type=["numericColumn"])
        builder.configure_column("second_repeat", header_name="Second Repeat", type=["numericColumn"])
        builder.configure_column("distance", header_name="Distance", type=["numericColumn"])
        builder.configure_column("mutation_rate", header_name="Mutation Rate",
                    type=["numericColumn"], valueFormatter="x.toExponential(2)")
        builder.configure_column("annotations", header_name="Annotations", tooltipField="annotations")
        builder.configure_column("predid", hide = True)
        builder.configure_column("annotationobjects", hide = True)

        grid_options = builder.build()
        self._srs_webapp_state = grid_options

    def update_srs_session(self, callbackobj):
        self.refresh_last_shown()
        state = callbackobj.grid_response["gridState"]
        new_selection = state.get('rowSelection', [])

        if self.last_srs_selections is not None and self.last_srs_selections == new_selection:
            return
        if self.last_srs_selections is None:
            self.last_srs_selections = []
        if self.last_srs_selections == new_selection:
            return

        self.counter += 1
        self.topkey = f"toptable{self.counter}"

        to_drop = [int(x) for x in self.last_srs_selections if x not in new_selection]
        to_add = [int(x) for x in new_selection if x not in self.last_srs_selections]


        self._update_general_table(self._filtered_srss, to_drop, to_add)

        self.last_srs_selections = new_selection

    # RMD Table Functions
    @property
    def rmd_webapp_table(self):
        """Streamlit aggrid table representing rmd data"""

        pandas_conversion = self._filtered_rmds.to_pandas()
        return AgGrid(pandas_conversion,
                            gridOptions=self._rmd_webapp_state,
                            height=500,
                            fit_columns_on_grid_load=True,
                            allow_unsafe_jscode = True,
                            update_mode = GridUpdateMode.SELECTION_CHANGED,
                            key=self.rmdkey,
                            callback=self.update_rmd_session,
                            reload_data=False)

    def rebuild_rmd_table(self):
        cell_hover_handler = JsCode("""
                    function(params) {
                        // debug
                        const clickedColumn = params.column.colId;
                        const clickedRowIndex = params.rowIndex;
                        const clickedValue = params.node.data[clickedColumn];

                        const predidValue = params.node.data["predid"];

                        // Display information about the click
                        const message = `You hovered on row ${clickedRowIndex}, column ${clickedColumn}, value is ${predidValue}`;
                        console.log(message);

                        window.parent.postMessage(
                            {type: 'cellMouseOver', predid: predidValue}, '*'
                            );

                    }
                    """)
        js_hover_handler = """

                    """

        preselected_indices = self.get_shown_predictions(self._filtered_rmds)
        preselected_indices = [str(index) for index in preselected_indices]
        if self.last_rmd_selections is None:
            self.last_rmd_selections = preselected_indices

        pandas_conversion = self._filtered_rmds.to_pandas()
        builder = GridOptionsBuilder.from_dataframe(pandas_conversion)
        builder.configure_selection(selection_mode='multiple', use_checkbox= True, pre_selected_rows= preselected_indices)
        builder.configure_grid_options(onCellMouseOver=cell_hover_handler)
        builder.configure_column("repeat", header_name="Sequence", tooltipField="repeat")
        builder.configure_column("repeat_len", header_name="Repeat Length", type=["numericColumn"])
        builder.configure_column("first_repeat", header_name="First Repeat", type=["numericColumn"])
        builder.configure_column("second_repeat", header_name="Second Repeat", type=["numericColumn"])
        builder.configure_column("distance", header_name="Distance", type=["numericColumn"])
        builder.configure_column("mutation_rate", header_name="Mutation Rate",
                    type=["numericColumn"], valueFormatter="x.toExponential(2)")
        builder.configure_column("annotations", header_name="Annotations", tooltipField="annotations")
        builder.configure_column("predid", hide = True)
        builder.configure_column("annotationobjects", hide = True)

        grid_options = builder.build()
        self._rmd_webapp_state = grid_options

    def update_rmd_session(self, callbackobj):
        self.refresh_last_shown()
        state = callbackobj.grid_response["gridState"]
        new_selection = state.get('rowSelection', [])

        if self.last_rmd_selections is not None and self.last_rmd_selections == new_selection:
            return
        if self.last_rmd_selections is None:
            self.last_rmd_selections = []
        if self.last_rmd_selections == new_selection:
            return

        self.counter += 1
        self.topkey = f"toptable{self.counter}"

        to_drop = [int(x) for x in self.last_rmd_selections if x not in new_selection]
        to_add = [int(x) for x in new_selection if x not in self.last_rmd_selections]


        self._update_general_table(self._filtered_rmds, to_drop, to_add)

        self.last_rmd_selections = new_selection


    def annotation_coverage(self, annotations):
        annotation_objects = self._unique_annotations.filter(pl.col("annotationobjexpanded_names").is_in(annotations))
        annotation_objects = annotation_objects.select(["left_bound", "right_bound"])

        coverage = []
        for row in annotation_objects.iter_rows(named=True):
            for i, occupied_area in enumerate(coverage):
                if occupied_area[0] <= row['left_bound'] and occupied_area[1] >= row['right_bound']:
                    # Entirely inside
                    break
                elif occupied_area[0] <= row['left_bound'] and row['left_bound'] <= occupied_area[1] <= row['right_bound']:
                    coverage[i][0] = row['left_bound']
                    break
                elif  row['left_bound'] <= occupied_area[0] <= row['right_bound'] and occupied_area[1] >= row['right_bound']:
                    coverage[i][1] = row['right_bound']
                    break
            else:
                # entirely outside
                coverage.append((row['left_bound'], row['right_bound']))
        base_coverage = 0
        for region in coverage:
            base_coverage += region[1] - region[0] + 1
        return base_coverage


    def get_shown_predictions(self, df):
        selected_rows = df.with_row_index().with_columns(
            pl.when(pl.col("predid").is_in(self._shown_predictions))
            .then(pl.lit(True))
            .otherwise(pl.lit(False))
            .alias("show")
        ).filter(pl.col("show") == True).select("index").unique().rows()
        selected_rows = [x for xs in selected_rows for x in xs] # Unnest
        return selected_rows

    def _update_general_table(self, dataframe, to_drop, to_add):
        """Internal function. Updates shown predictions based on those selected in table"""
        to_drop = dataframe.with_row_index(
                    ).filter(pl.col("index").is_in(to_drop)
                    ).select('predid').rows()
        to_drop = [x for xs in to_drop for x in xs] # Unnest

        to_add = dataframe.with_row_index(
                    ).filter(pl.col("index").is_in(to_add)
                    ).select('predid').rows()
        to_add = [x for xs in to_add for x in xs] # Unnest

        for item in to_drop:
            try:
                self._shown_predictions.remove(item)
            except ValueError:
                pass # Eronious removal
        for item in to_add:
            if item not in self._shown_predictions:
                self._shown_predictions.append(item)

    def refresh_last_shown(self):
        if not self.refreshed:
            #top
            preselected_indices = self.get_shown_predictions(self._filtered_top)
            preselected_indices = [str(index) for index in preselected_indices]
            self.last_top_selections = preselected_indices
            #ssrs
            preselected_indices = self.get_shown_predictions(self._filtered_ssrs)
            preselected_indices = [str(index) for index in preselected_indices]
            self.last_ssr_selections = preselected_indices
            #srss
            preselected_indices = self.get_shown_predictions(self._filtered_srss)
            preselected_indices = [str(index) for index in preselected_indices]
            self.last_srs_selections = preselected_indices
            #rmd
            preselected_indices = self.get_shown_predictions(self._filtered_rmds)
            preselected_indices = [str(index) for index in preselected_indices]
            self.last_rmd_selections = preselected_indices
            self.refreshed = True



from Bio.SeqRecord import SeqRecord
from Bio.SeqFeature import SimpleLocation, CompoundLocation, SeqFeature
from types import NoneType

import polars as pl
from ..webapp.vis_utils import eval_top

from ..pipeline.features import seqfeature_hash

from ..pipeline.primary_pipeline import predict
from ..pipeline.post_process import post_process

class EFMSequence(SeqRecord):
    """SeqRecord child class with equality handling and prediction methods"""
    def __init__(self, seqrecord: SeqRecord, is_circular = False, originhash = None):
        super().__init__(seq=seqrecord.seq,
            id=seqrecord.id,
            name=seqrecord.name,
            description=seqrecord.description,
            dbxrefs = seqrecord.dbxrefs,
            features = seqrecord.features,
            annotations = seqrecord.annotations,
            letter_annotations = seqrecord.letter_annotations)
        self.seq = seqrecord.seq
        self.formatted_name = None
        self.is_circular = is_circular

        self._predicted = False
        self._top = None
        self._ssrs = None
        self._srss = None
        self._rmds = None

        self._unique_annotations = {}
        self._originhash = originhash

        self._filtered_ssrs = None
        self._filtered_srss = None
        self._filtered_rmds = None
        self._filtered_top = None
        self._last_filters = []

        self._plotted_predictions = []
        self._ssr_webapp_state = None


    @property
    def seq(self):
        return self._seq

    @seq.setter
    def seq(self, seq):
        self._ssrs = None
        self._srss = None
        self._rmds = None
        self._predicted = False
        self._seq = seq

    @property
    def predicted(self):
        return self._predicted

    @property
    def top(self):
        if self._top is None:
            raise ValueError("No predictions calculated. Run call_predictions() first.")
        return self._top

    @property
    def ssrs(self):
        if self._ssrs is None:
            raise ValueError("No SSRs calculated. Run call_predictions() first.")
        return self._ssrs

    @property
    def srss(self):
        if self._srss is None:
            raise ValueError("No SRSs calculated. Run call_predictions() first.")
        return self._srss

    @property
    def rmds(self):
        if self._rmds is None:
            raise ValueError("No RMDs calculated. Run call_predictions() first.")
        return self._rmds

    def call_predictions(self, strategy):
        seq = str(self.seq.strip("\n").upper().replace("U", "T"))
        ssr_df, srs_df, rmd_df = predict(seq, strategy, self.is_circular)
        ssr_df, srs_df, rmd_df = post_process(ssr_df,
                                                             srs_df,
                                                             rmd_df,
                                                             self,
                                                             self.is_circular)
        self._ssrs = ssr_df.with_columns(
            pl.concat_str(pl.col(['repeat', 'repeat_len', "start", "count", "mutation_rate"])).hash().cast(pl.String).alias('predid')
        )
        self._srss = srs_df.with_columns(
            pl.concat_str(pl.col(['repeat', 'repeat_len', "first_repeat", "second_repeat", "distance", "mutation_rate"])).hash().cast(pl.String).alias('predid')
        )
        self._rmds = rmd_df.with_columns(
            pl.concat_str(pl.col(['repeat', 'repeat_len', "first_repeat", "second_repeat", "distance", "mutation_rate"])).hash().cast(pl.String).alias('predid')
        )
        self._top = eval_top(self._ssrs, self._srss, self._rmds)
        self._predicted = True

    def same_origin(self, other):
        if not self._originhash or not other._filehash:
            return False
        elif self._originhash != other._filehash:
            return False
        else:
            return True

def sequence_to_features_df(sequence, circular=True):
    """Takes genbank annotations and turns them into a polars dataframe"""
    features = sequence.features
    seqlen = len(sequence)

    def get_feature_bounds(feature_location):
        if isinstance(feature_location, SimpleLocation):
            return feature_location.start, feature_location.end
        elif isinstance(feature_location, CompoundLocation):
            if feature_location.parts[-1].end < feature_location.parts[0].start:
                end = len(sequence) + feature_location.parts[-1].end
            else:
                end = feature_location.parts[-1].end
            return feature_location.parts[0].start, end

    if not circular:
        # Look for compound wraparounds and break them into simple features
        newfeatures = []
        deletedfeatures = []
        for feature in features:
            if not isinstance(feature.location, CompoundLocation):
                continue

            # Check whether the compound feature is actually a wraparound
            wraparound_part_index = None
            last_part_start = None
            rightmost_part = None
            for i, part in enumerate(feature.location.parts):
                if rightmost_part != None and part.start < last_part_start:
                    wraparound_part_index = i
                if rightmost_part == None:
                    rightmost_part = i
                    last_part_start = part.start
            if wraparound_part_index is None:
                continue

            # If it is a wraparound, break it into two features
            leftsplit_locations = feature.location.parts[:wraparound_part_index]
            if len(leftsplit_locations) > 1:
                leftsplit_locations = CompoundLocation(leftsplit_locations)
            else:
                leftsplit_locations = leftsplit_locations[0]
            newleftfeature = SeqFeature(leftsplit_locations, feature.type, qualifiers=feature.qualifiers)


            rightsplit_locations = feature.location.parts[wraparound_part_index:]
            if len(rightsplit_locations) > 1:
                rightsplit_locations = CompoundLocation(rightsplit_locations)
            else:
                rightsplit_locations = rightsplit_locations[0]
            newrightfeature = SeqFeature(rightsplit_locations, feature.type, qualifiers=feature.qualifiers)

            newfeatures.append(newleftfeature)
            newfeatures.append(newrightfeature)
            deletedfeatures.append(feature)

        for feature in deletedfeatures:
            features.remove(feature)
        features.extend(newfeatures)

    df = pl.DataFrame([(feature.type,
                        get_feature_bounds(feature.location),
                        feature.qualifiers.get("label", ["unlabeled"])[0],
                        seqfeature_hash(feature)) for feature in features],
        schema=['type', 'loc', 'annotations', 'annotationobjects'],
        orient="row")

    print(df)

    # expand out loc
    df = df.with_columns(pl.col("loc").list.to_struct(fields=['left_bound', 'right_bound'])).unnest("loc")
    df = df.with_columns(
        pl.col("left_bound").cast(pl.Int32),
        pl.col("right_bound").cast(pl.Int32)
    )

    return df



import polars as pl

def filter_ssrs(ssr_dataframe, seq_len, circular):
    if len(ssr_dataframe) > 0:
        ssr_dataframe = (
            ssr_dataframe.lazy()
            # Filter based on SSR definition
            .filter(
                (pl.col("repeat_len") >= 3)
                .and_(pl.col("count") >= 3)
                .or_((pl.col("repeat_len") <= 2).and_(pl.col("count") >= 4))
            )   #! EFM1 is bugged with repeat_ len>= 3 and <= 2 bp
        ).collect()

        ssr_dataframe = (
        ssr_dataframe.sort([pl.col("start"), pl.col("repeat_len")], descending=[False, True])
            .with_columns(
                # end is the last base pair of the SSR
                (pl.col("start") + (pl.col("repeat_len")*pl.col("count"))-1).alias("end")
            )
            .with_columns(
                # if circular, creates modified start and stop positions
                pl.when(circular)
                .then(
                    pl.when((pl.col("end")) >= seq_len)
                    .then((pl.col("start")-seq_len).alias("modified_start"))
                    .otherwise(pl.col("start").alias("modified_start"))
                    )
                .otherwise(pl.col("start").alias("modified_start")),
                pl.when(circular)
                .then(
                    pl.when((pl.col("end")) >= seq_len)
                    .then((pl.col("end")-seq_len).alias("modified_end"))
                    .otherwise(pl.col("end").alias("modified_end"))
                    )
                .otherwise(pl.col("end").alias("modified_end"))
            )
            .with_row_index()
            )
        
        joined = ssr_dataframe.join(ssr_dataframe, how="cross")

        # remove rows compared to itself
        joined = joined.filter(pl.col("index") != pl.col("index_right"))
        if(not circular):
            joined = (
            joined.filter(
                    # these repeats are fully contained in another repeat, and has less count
                    (
                        (pl.col("start") >= pl.col("start_right")) &
                        (pl.col("end") <= pl.col("end_right")) &
                        (pl.col("count") <= pl.col("count_right"))
                    ) |
                    # these repeats are alternate versions of other SSRs
                    (
                        (pl.col("start") - pl.col("start_right") == pl.col("end") - pl.col("end_right")) &
                        ((pl.col("start") - pl.col("start_right")).abs() < pl.col("repeat_len")) &
                        # want to keep the first version 
                        (pl.col("start") > pl.col("start_right"))
                    )
                )
            )
        else:
            joined = (
            joined.filter(
                    # these repeats are fully contained in another repeat, and has less count
                    (
                        (
                            # use either modified start or modified end for repeats that wrap around
                            ((pl.col("start") >= pl.col("start_right")) &
                            (pl.col("end") <= pl.col("end_right"))
                            ) |
                            ((pl.col("start") >= pl.col("modified_start_right")) &
                            (pl.col("end") <= pl.col("modified_end_right"))
                            ) |
                            ((pl.col("modified_start") >= pl.col("start_right")) &
                            (pl.col("modified_end") <= pl.col("end_right"))
                            ) |
                            ((pl.col("modified_start") >= pl.col("modified_start_right")) &
                            (pl.col("modified_end") <= pl.col("modified_end_right"))
                            )
                        ) &
                        (pl.col("count") <= pl.col("count_right"))
                    ) |
                    # these repeats are alternate versions of other SSRs
                    (
                        (pl.col("start") - pl.col("start_right") == pl.col("end") - pl.col("end_right")) &
                        ((pl.col("start") - pl.col("start_right")).abs() < pl.col("repeat_len")) &
                        # want to keep the first version 
                        (pl.col("start") > pl.col("start_right"))
                    )
                )
            )

        removed_indices = joined["index"].unique()

        ssr_dataframe = (
            ssr_dataframe
            .filter(
                ~(pl.col("index").is_in(removed_indices))
            )
        .select(["repeat", "repeat_len", "start", "count"])
        # fix start values
        .with_columns(
            pl.when(pl.col("start") < 0)
            .then(pl.col("start") + seq_len)
            .otherwise(pl.col("start"))
            )
        )

    return ssr_dataframe


def filter_direct_repeats(rmd_dataframe, srs_dataframe, seq_len, ssr_dataframe, circular):
    # Delete redundant SRS repeats

    # label RMDs and SRSs and combine into 1 df
    rmd_dataframe = rmd_dataframe.with_columns(pl.lit("RMD").alias("type"))

    srs_dataframe = srs_dataframe.with_columns(pl.lit("SRS").alias("type"))
    combined_dataframe = pl.concat([srs_dataframe, rmd_dataframe])



    #filter combined dataframe
    combined_dataframe = (
        combined_dataframe
        .filter(
            pl.col("repeat_len") > 4
        )
        # filter out circular overlapping repeats
        .with_columns(
            (pl.col("first_repeat") + pl.col("repeat_len")).alias("left_end"),
            (pl.col("second_repeat") + pl.col("repeat_len")).alias("right_end")
        )
        .with_columns(
            pl.when(circular)
            .then(
                pl.when((pl.col("right_end")) >= seq_len)
                .then(pl.col("right_end") - seq_len)
                .otherwise(pl.col("right_end"))
            )
            .otherwise(pl.col("right_end"))
        )
        # noncircular overlapping repeats are covered by pairwise approach already
        .filter(
            ~(
                (pl.col("right_end") > pl.col("first_repeat")) &
                (pl.col("right_end") < pl.col("left_end"))
            )
        )
        .sort(["first_repeat", "repeat_len"], descending=[False, True])
        .group_by(pl.col("first_repeat"), pl.col("repeat"))
        .agg(
            pl.col("second_repeat"),
            pl.first("repeat_len"),
            pl.col("distance"),
            pl.col("type")
        )
        .sort(["first_repeat", "repeat_len"], descending=[False, True])
        .with_columns(
            pl.col("first_repeat").shift(1).alias("last_first_repeat"),
            pl.col("second_repeat").shift(1).list.unique().alias("last_second_repeat"),
            pl.col("repeat_len").shift(1).alias("last_len")
        )
        .explode(["second_repeat", "distance", "type"])
        .with_row_index()
    )


    # Create a list of indices that should be deleted
    # Delete instances of smaller repeats that are in bigger repeats (ex. 2 copies of "AAGTCAT" and 3 copies of "AAGTCA". Delete the instance of "AAGTCA" that 
    # corresponds to the "AAGTCAT" repeat and keep the other 2 pairwise repeats)
    filter_out = (
        combined_dataframe
        .filter(
            (pl.col("first_repeat") == pl.col("last_first_repeat")) &
            (pl.col("last_second_repeat").list.contains(pl.col("second_repeat"))) &
            (pl.col("repeat_len") < pl.col("last_len"))
            )
        .select("index").to_series().to_list()
    )
    

    # Create another list of indices that should be deleted
    # Delete shorter versions of the same repeat that start at different positions
    filter_out_2 = (
        combined_dataframe
        .group_by("first_repeat", maintain_order=True)
        .agg(
            pl.col("repeat"),
            pl.col("second_repeat"),
            pl.first("repeat_len"),
            pl.col("distance"),
            pl.col("type"), 
            pl.first("last_first_repeat"), 
            pl.first("last_second_repeat"),
            pl.first("last_len"), 
            pl.col("index")
        )
        .with_columns(
            pl.col("first_repeat").shift(1).alias("last_first_repeat"),
            pl.col("second_repeat").shift(1).list.unique().alias("last_second_repeat"),
            pl.col("repeat_len").shift(1).alias("last_len")
        )
        .explode(["repeat", "second_repeat", "distance", "type", "index"])
        .with_columns(
            (pl.col("first_repeat") - pl.col("last_first_repeat")).alias("difference")
        )
        .with_columns(
            (pl.col("first_repeat") - pl.col("difference")).alias("adjusted_first_repeat"),
            (pl.col("second_repeat") - pl.col("difference")).alias("adjusted_second_repeat")
        )
        .filter(
            (pl.col("adjusted_first_repeat") == pl.col("last_first_repeat")) &
            (pl.col("last_second_repeat").list.contains(pl.col("adjusted_second_repeat"))) &
            (pl.col("repeat_len") <= pl.col("last_len"))
        )
        .select("index").to_series().to_list()
        )

    # Filter out repetas with indices in either of the lists
    filtered_df = combined_dataframe.filter(
        ~(
            (pl.col("index").is_in(filter_out)) |
            (pl.col("index").is_in(filter_out_2))
        )
    )


    # filter out SRS nested fully inside SSRs
    # if statement needed because cross join with an empty df creates an empty df
    if ssr_dataframe.height > 0:
        ssr_dataframe = (
            ssr_dataframe
            .select(
                # start is the 1st bp of SSR
                pl.col("start"),
                # end is the last bp of SSR
                (pl.col("start")+(pl.col("count")*pl.col("repeat_len")) - 1).alias("end"),
                pl.col("repeat_len"),
                pl.col("count")
            )
            .with_columns(
                pl.when(circular)
                .then(
                    pl.when(pl.col("end") >= seq_len)
                    .then(True)
                    .otherwise(False)
                    .alias("wraparound")
                )
                .otherwise(False).alias("wraparound")
            )
            .with_columns(
                (pl.col("repeat_len")*pl.col("count")).alias("ssr_length"),
                pl.when(pl.col("wraparound") == True)
                .then(pl.col("end")-seq_len)
                .otherwise(pl.col("end"))
            )
            .select("start", "end", "ssr_length", "wraparound")
        )

        ssr_ranges = list(ssr_dataframe.select([pl.col("start"), pl.col("end"), pl.col("wraparound")]).iter_rows())
        

        # Function to check 'nested' condition for a single row against all tuples
        def check_nested(row):
            first_repeat = row["first_repeat"]
            second_repeat = row["second_repeat"]
            repeat_len = row["repeat_len"]

            nested_values = []
            for start, end, wraparound in ssr_ranges:
                if wraparound:
                    condition = not (
                        ((first_repeat < start) & ((first_repeat + repeat_len - 2) > end)) |
                        ((second_repeat < start) & ((second_repeat + repeat_len - 2) > end))
                    )
                else:
                    condition = (first_repeat >= start) & ((second_repeat + repeat_len - 1) <= end)
                
                nested_values.append(condition)

            return any(nested_values)  # If any tuple results in True, set nested=True


        # Apply function to each row
        filtered_df = (
            filtered_df.with_columns(
                pl.struct(["first_repeat", "second_repeat", "repeat_len"])
                .map_elements(check_nested, return_dtype=pl.Boolean)
                .alias("nested")
            )
            .filter(~pl.col("nested"))  # Filter out rows where nested=True
        )

    filtered_df = filtered_df.select("repeat", "repeat_len", "first_repeat", "second_repeat", "distance", "type")

    # split back into rmd_dataframe and srs_dataframe
    rmd_dataframe = filtered_df.filter(pl.col("type") == "RMD").drop("type")
    srs_dataframe = filtered_df.filter(pl.col("type") == "SRS").drop("type")

    return rmd_dataframe, srs_dataframe


# No longer necessary filters (already covered by pairwise approach)
# - Delete rows with overlapping repeats and only 2 occurrences (not a real repeat)
